# Output backends for sigmac
# Copyright 2021 Devo, Inc.
# Author: Eduardo Ocete <eduardo.ocete@devo.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from .base import SingleTextQueryBackend
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier
from sigma.parser.condition import SigmaAggregationParser
from sigma.parser.exceptions import SigmaParseError

class DevoBackend(SingleTextQueryBackend):
    """Converts Sigma rule into Devo query."""
    identifier = "devo"
    active = True

    andToken = " and "                            # Token used for linking expressions with logical AND
    orToken = " or "                              # Same for OR
    notToken = " not "                            # Same for NOT
    subExpression = "(%s)"                        # Syntax for subexpressions, usually parenthesis around it. %s is inner expression
    listExpression = "%s"                         # Syntax for lists, %s are list items separated with listSeparator
    listSeparator = ", "                          # Character for separation of list items
    valueExpression = "\"%s\""                    # Expression of values, %s represents value
    intValueExpression = "%s"                     # Expression of int values, %s represents value
    nullExpression = "isnull(%s)"                 # Expression of queries for null values or non-existing fields. %s is field name
    notNullExpression = "isnotnull(%s)"           # Expression of queries for not null values. %s is field name
    mapExpression = "%s = %s"                     # Syntax for field/value conditions. First %s is fieldname, second is value
    mapMulti = "has(%s, %s)"                      # Syntax for field/value conditions. First %s is fieldname, second is value
    mapWildcard = "matches(%s, nameglob(%s))"     # Syntax for globbing conditions
    mapRe = "matches(%s, %s)"                     # Syntax for regex conditions that already were transformed by SigmaRegularExpressionModifier
    mapContains = "toktains(%s, %s, true, true)"  # Syntax for token value searches
    mapListValueExpression = "%s or %s"           # Syntax for field/value condititons where map value is a list
    mapFullTextSearch = "weaktoktains(raw, \"%s\", true, true)"  # Expression for full text searches
    typedValueExpression = {
        SigmaRegularExpressionModifier: "re(\"%s\")",  # Syntax for regular expressions
    }

    # \   -> \\
    # \*  -> \*
    # \\* -> \\*
    reEscape = re.compile('("|(?<!\\\\)\\\\(?![*?\\\\]))')
    derivedField = re.compile('^select .* as (.+)$')
    derivedFieldSet = set()
    hasMulticondition = False

    def __init__(self, sigmaconfig, options):
        super().__init__(sigmaconfig)
        # Default table name. It is replaced based on the config file
        self.table = "sourcetable"

    def generateANDNode(self, node):
        generated = []
        for val in node:
            if self.requireFTS(val):
                generated.append(self.generateFTS(val))
            else:
                generated.append(self.generateNode(val))

        filtered = [g for g in generated if g is not None]
        if filtered:
            return self.andToken.join(filtered)
        else:
            return None

    def generateORNode(self, node):
        generated = []
        for val in node:
            if self.requireFTS(val):
                generated.append(self.generateFTS(val))
            else:
                generated.append(self.generateNode(val))

        filtered = [g for g in generated if g is not None]
        if filtered:
            return self.orToken.join(filtered)
        else:
            return None

    def generateNOTNode(self, node):
        if self.requireFTS(node.item):
            generated = self.generateFTS(node.item)
        else:
            generated = self.generateNode(node.item)

        if generated is not None:
            return self.notToken + generated
        else:
            return None

    def generateSubexpressionNode(self, node):
        generated = self.generateNode(node.items)
        if generated:
            return self.subExpression % generated
        else:
            return None

    def generateListNode(self, node):
        if not set([type(value) for value in node]).issubset({str, int}):
            raise TypeError("List values must be strings or numbers")
        return self.listExpression % (self.listSeparator.join([self.generateNode(value) for value in node]))

    def generateMapItemNode(self, node):
        fieldname, value = node
        transformed_fieldname = self.fieldNameMapping(fieldname, value)

        if not value:
            # Handle value == None
            return self.generateNULLValueNode(transformed_fieldname)

        has_startswith = self.generateNode(value).startswith("\"*")
        has_endswith = self.generateNode(value).endswith("*\"")
        has_contains = has_startswith and has_endswith and len(self.generateNode(value)) > 3  # Covers "*" case

        if type(value) == SigmaRegularExpressionModifier:
            return self.mapRe % (transformed_fieldname, self.generateNode(value))
        elif type(value) == list:
            if has_contains:
                return self.subExpression % self.andToken.join(self.mapContains % (transformed_fieldname, self.generateNode(val[1:-1])) for val in value)
            elif has_startswith or has_endswith:
                return self.generateMapItemListNode(transformed_fieldname, value)
            else:
                return self.mapMulti % (transformed_fieldname, self.generateNode(value))
        elif type(value) in (str, int):
            if has_contains:
                return self.mapContains % (transformed_fieldname, self.generateNode(value[1:-1]))
            elif has_startswith or has_endswith:
                return self.mapWildcard % (transformed_fieldname, self.generateNode(value))
            else:
                return self.mapExpression % (transformed_fieldname, self.generateNode(value))
        else:
            raise TypeError("Devo backend does not support map values of type " + str(type(value)))

    def generateMapItemListNode(self, key, value):
        return "(" + (" or ".join([self.mapWildcard % (key, self.generateValueNode(item)) for item in value])) + ")"

    def generateValueNode(self, node):
        if type(node) == int:
            return self.intValueExpression % int(node)
        return self.valueExpression % (self.cleanValue(node))

    def generateNULLValueNode(self, fieldname):
        return self.nullExpression % fieldname

    def generateNotNULLValueNode(self, fieldname):
        return self.notNullExpression % fieldname

    def generateTypedValueNode(self, node):
        try:
            return self.typedValueExpression[type(node)] % (self.cleanValue(str(node)))
        except KeyError:
            raise NotImplementedError("Type modifier '{}' is not supported by backend".format(node.identifier))

    def generateFTS(self, value):
        return self.mapFullTextSearch % self.cleanValue(value)

    def requireFTS(self, value):
        return isinstance(value, str) or isinstance(value, int) or isinstance(value, list)

    def fieldNameMapping(self, field, value):
        # Handle derived fields
        matched = self.derivedField.search(field)
        if matched:
            self.derivedFieldSet.add(field)
            return matched.group(1)
        return field

    def generateAggregation(self, agg, where_clause):
        if not agg:
            return self.table, where_clause
        # Near operator not supported yet
        if agg.aggfunc == SigmaAggregationParser.AGGFUNC_NEAR:
            raise NotImplementedError("The 'near' aggregation operator is not implemented for the %s backend" % self.identifier)
        if (agg.aggfunc == SigmaAggregationParser.AGGFUNC_COUNT or
                agg.aggfunc == SigmaAggregationParser.AGGFUNC_MAX or
                agg.aggfunc == SigmaAggregationParser.AGGFUNC_MIN or
                agg.aggfunc == SigmaAggregationParser.AGGFUNC_SUM or
                agg.aggfunc == SigmaAggregationParser.AGGFUNC_AVG):

            if agg.groupfield and self.timeframe == None:
                if self.hasMulticondition:
                    group_by = " group every - by subquery_link,{0}".format(self.fieldNameMapping(agg.groupfield, None))
                else:
                    group_by = " group by {0}".format(self.fieldNameMapping(agg.groupfield, None))
            elif agg.groupfield and self.timeframe != None:
                if self.hasMulticondition:
                    group_by = " group every {} by subquery_link,{0}".format(self.timeframe,self.fieldNameMapping(agg.groupfield, None))
                else:
                    group_by = " group by {} every {}".format(self.fieldNameMapping(agg.groupfield, None), self.timeframe)
            else:
                group_by = ""

            if agg.aggfield:
                select = "{}({}) as agg".format(agg.aggfunc_notrans, self.fieldNameMapping(agg.aggfield, None))
            else:
                if agg.aggfunc == SigmaAggregationParser.AGGFUNC_COUNT:
                    select = "{}(*) as agg".format(agg.aggfunc_notrans)
                else:
                    raise SigmaParseError("For {} aggregation a fieldname needs to be specified".format(agg.aggfunc_notrans))

            if self.derivedFieldSet:
                derivedFieldsStr = " {}".format(" ".join(self.derivedFieldSet))
            else:
                derivedFieldsStr = ""

            if self.hasMulticondition:
                link_select = ' select "link" as subquery_link'
            else:
                link_select = ""

            temp_table = "from {}{} where {}{}{} select {}".format(self.table, derivedFieldsStr, where_clause, link_select, group_by, select)
            agg_condition = "agg {} {}".format(agg.cond_op, agg.condition)

            return temp_table, agg_condition

        raise NotImplementedError("{} aggregation not implemented in Devo Backend".format(agg.aggfunc_notrans))

    def generateQuery(self, parsed):
        if self.requireFTS(parsed.parsedSearch):
            result = self.generateFTS(parsed.parsedSearch)
        else:
            result = self.generateNode(parsed.parsedSearch)
        if parsed.parsedAgg:
            fro, whe = self.generateAggregation(parsed.parsedAgg, result)
            return "{} where {} select *".format(fro, whe)

        if self.derivedFieldSet:
            derivedFieldsStr = " {}".format(" ".join(self.derivedFieldSet))
        else:
            derivedFieldsStr = ""

        if self.hasMulticondition:
            select = 'select "link" as subquery_link'
        else:
            select = "select *"

        return "from {}{} where {} {}".format(self.table, derivedFieldsStr, result, select)

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        self.derivedFieldSet = set()
        if sigmaparser.get_logsource() and sigmaparser.get_logsource().index:
            self.table = sigmaparser.get_logsource().index[0]
        else:
            self.table = "sourcetable"

        try:
            self.timeframe = sigmaparser.parsedyaml['detection']['timeframe']
        except:
            self.timeframe = None
            pass
        
        if len(sigmaparser.condparsed) > 1:
            self.hasMulticondition = True
        else:
            self.hasMulticondition = False

        results = []
        for parsed in sigmaparser.condparsed:
            # Multi condition rules are not supported yet, only the first one will be processed
            query = self.generateQuery(parsed)
            before = self.generateBefore(parsed)
            after = self.generateAfter(parsed)

            result = ""
            if before is not None:
                result = before
            if query is not None:
                result += query
            if after is not None:
                result += after

            results.append(result)

        if self.hasMulticondition:
            prefix = 'from siem.logtrust.alert.info select "link" as subquery_link group every 24h by subquery_link where '
            suffix = " select *"
            for i in range(len(results)):
                results[i] = "subquery_link in ( " + results[i]
                results[i] += ")"

            body = " or ".join(results)

            return prefix + body + suffix

        return results[0]



# Output backends for sigmac 
# Copyright 2021 HAWK.io

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import re
import sigma
import json
import uuid
import re
from sigma.parser.modifiers.base import SigmaTypeModifier
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier
from .base import SingleTextQueryBackend
from .mixins import MultiRuleOutputMixin




class HAWKBackend(SingleTextQueryBackend):
    """Converts Sigma rule into HAWK search"""
    identifier = "hawk"
    mitre_json_url = "https://portal.hawk.io:8080/API/1.1/analytics/attack"
    active = True
    config_required = False
    default_config = ["sysmon", "hawk"]
    reEscape = re.compile('(")')
    logname = None
    reClear = None
    andToken = " , "
    orToken = " , "
    subExpression = "{\"id\": \"and\", \"key\": \"And\", \"children\": [%s] }"
    listExpression = "%s"
    listSeparator = " "
    valueExpression = "%s"
    keyExpression = "%s"
    nullExpression = "%s = null"
    notNullExpression = "%s != null"
    mapExpression = "%s=%s"
    mapListsSpecialHandling = True
    aql_database = "events"

    def cleanKey(self, key):
        if key == None:
            return ""
        return self.snake_case( self.sigmaparser.config.get_fieldmapping(key).resolve_fieldname(key, self.sigmaparser) )

    def cleanValue(self, value):
        """Remove quotes in text"""
        return value

    def generateNode(self, node, notNode=False):
        #print(type(node))
        #print(node)
        #print("Not: ", notNode)
        if type(node) == sigma.parser.condition.ConditionAND:
            return self.generateANDNode(node, notNode)
        elif type(node) == sigma.parser.condition.ConditionOR:
            #print("OR NODE")
            #print(node)
            return self.generateORNode(node, notNode)
        elif type(node) == sigma.parser.condition.ConditionNOT:
            #print("NOT NODE")
            #print(node)
            return self.generateNOTNode(node)
        elif type(node) == sigma.parser.condition.ConditionNULLValue:
            return self.generateNULLValueNode(node, notNode)
        elif type(node) == sigma.parser.condition.ConditionNotNULLValue:
            return self.generateNotNULLValueNode(node)
        elif type(node) == sigma.parser.condition.NodeSubexpression:
            #print(node)
            return self.generateSubexpressionNode(node, notNode)
        elif type(node) == tuple:
            #print("TUPLE: ", node)
            return self.generateMapItemNode(node, notNode)
        elif type(node) in (str, int):
            nodeRet = {"key": "",  "description": "", "class": "column", "return": "str", "args": { "comparison": { "value": "=" }, "str": { "value": "5", "regex": "true" } } }
            #key = next(iter(self.sigmaparser.parsedyaml['detection'])) 
            key = "payload"

            #nodeRet['key'] = self.cleanKey(key).lower()
            nodeRet['key'] = key

            #print(node)
            #print("KEY: ", key)
            # they imply the entire payload
            nodeRet['description'] = key
            nodeRet['rule_id'] = str(uuid.uuid4())
            value = self.generateValueNode(node, False).replace("*", "EEEESTAREEE")
            if value[-2:] == "\\\\":
                value = value[:-2]
            value = re.escape(value)
            value = value.replace("EEEESTAREEE", ".*")
            endsWith = False
            startsWith = False
            if value[0:2] == ".*":  
                value = value[2:]
                endsWith = True
            if value[-2:] == ".*":
                value = value[:-2]
                startsWith = True

            if endsWith and not startsWith:
                nodeRet['args']['str']['value'] = value + "$"
            elif startsWith and not endsWith:
                nodeRet['args']['str']['value'] = "^" + value
            else:

                # custom, since we trim up  string size in log to save bytes
                if key == 'Provider_Name':
                    nodeRet['key'] = "product_name"
                    if type(value) is str and value[0:17] == 'Microsoft-Windows':
                        value = value[18:]

                nodeRet['args']['str']['value'] = value

            if notNode:
                nodeRet["args"]["comparison"]["value"] = "!="

            return nodeRet
        elif type(node) == list:
            return self.generateListNode(node, notNode)
        else:
            raise TypeError("Node type %s was not expected in Sigma parse tree" % (str(type(node))))

    def generateANDNode(self, node, notNode=False):
        ret = { "id" : "and", "key": "And", "children" : [ ] }
        generated = [ self.generateNode(val, notNode) for val in node ]
        filtered = [ g for g in generated if g is not None ]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            ret['children'] = filtered
            return ret
        else:
            return None

    def generateORNode(self, node, notNode=False):
        if notNode:
            ret = { "id" : "and", "key": "And", "children" : [ ] }
        else:
            ret = { "id" : "or", "key": "Or", "children" : [ ] }
        generated = [ self.generateNode(val, notNode) for val in node ]
        filtered = [ g for g in generated if g is not None ]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            ret['children'] = filtered

            return ret
        else:
            return None

    def generateSubexpressionNode(self, node, notNode=False):
        generated = self.generateNode(node.items, notNode)
        if 'len'in dir(node.items): # fix the "TypeError: object of type 'NodeSubexpression' has no len()"
            if len(node.items) == 1:
                # A sub expression with length 1 is not a proper sub expression, no self.subExpression required
                return generated
        if generated:
            return json.loads(self.subExpression % json.dumps(generated))
        else:
            return None

    def generateListNode(self, node, notNode=False):
        if not set([type(value) for value in node]).issubset({str, int}):
            raise TypeError("List values must be strings or numbers")
        result = [self.generateNode(value, notNode) for value in node]
        if len(result) == 1:
            # A list with length 1 is not a proper list, no self.listExpression required
            return result[0]
        return self.listExpression % (self.listSeparator.join(result))

    def generateNOTNode(self, node):
        generated = self.generateNode(node.item, True)
        return generated

    def generateMapItemNode(self, node, notNode=False):
        nodeRet = {"key": "",  "description": "", "class": "column", "return": "str", "args": { "comparison": { "value": "=" }, "str": { "value": 5 } } }
        if notNode:
            nodeRet["args"]["comparison"]["value"] = "!="
        nodeRet['rule_id'] = str(uuid.uuid4())
        key, value = node
        if self.mapListsSpecialHandling == False and type(value) in (str, int, list, bool) or self.mapListsSpecialHandling == True and type(value) in (str, int, bool):
            nodeRet['key'] = self.cleanKey(key).lower()
            nodeRet['description'] = key
            if key.lower() in ("logname","source"):
                self.logname = value
            if type(value) == str and "*" in value:

                if (nodeRet['key'] == 'correlation_username' or nodeRet['key'] == 'target_username'):
                    if 'NT AUTHORITY\\SYS' in value.upper():
                        value = value.replace('NT AUTHORITY\\SYSTEM', 'SYSTEM')
                    elif 'AUTHORI' in value.upper():
                        value = 'SYSTEM'

                value = value.replace("*", "EEEESTAREEE")
                value = re.escape(value)
                value = value.replace("EEEESTAREEE", ".*")
                endsWith = False
                startsWith = False
                if value[0:2] == ".*":  
                    value = value[2:]
                    endsWith = True
                if value[-2:] == ".*":
                    value = value[:-2]
                    startsWith = True
                if notNode:
                    nodeRet["args"]["comparison"]["value"] = "!="
                else:
                    nodeRet['args']['comparison']['value'] = "="
                if value[-2:] == "\\\\":
                    value = value[:-2]

                if endsWith and not startsWith:
                    nodeRet['args']['str']['value'] = value + "$"
                elif startsWith and not endsWith:
                    nodeRet['args']['str']['value'] = "^" + value
                else:
                    nodeRet['args']['str']['value'] = value

                nodeRet['args']['str']['regex'] = "true"
                # return "%s regex %s" % (self.cleanKey(key), self.generateValueNode(value, True))
                #return json.dumps(nodeRet)
                return nodeRet
            elif type(value) is str:
                if notNode:
                    nodeRet["args"]["comparison"]["value"] = "!="
                else:
                    nodeRet['args']['comparison']['value'] = "="

                # custom, since we trim up  string size in log to save bytes
                if key == 'Provider_Name':
                    nodeRet['key'] = "product_name"
                    if value[0:17] == 'Microsoft-Windows':
                        value = value[18:]

                if (nodeRet['key'] == 'correlation_username' or nodeRet['key'] == 'target_username'): 
                    if 'NT AUTHORITY\\SYS' in value.upper():
                        value = value.replace('NT AUTHORITY\\SYSTEM', 'SYSTEM')
                    elif 'AUTHORI' in value.upper():
                        value = 'SYSTEM'

                nodeRet['args']['str']['value'] = value
                # return json.dumps(nodeRet)
                return nodeRet
            elif type(value) is int:
                nodeRet['return'] = "int"
                nodeRet['args']['int'] = { "value" : value }
                if notNode:
                    nodeRet["args"]["comparison"]["value"] = "!="
                else:
                    nodeRet['args']['comparison']['value'] = "="
                del nodeRet['args']['str'] 
                #return self.mapExpression % (self.cleanKey(key), self.generateValueNode(value, True))
                #return json.dumps(nodeRet)
                return nodeRet
            elif type(value) is bool:
                nodeRet['return'] = "bool"
                nodeRet['args']['bool'] = { "value" : value }

                if notNode:
                    nodeRet["args"]["comparison"]["value"] = "!="
                else:
                    nodeRet['args']['comparison']['value'] = "="
                del nodeRet['args']['str'] 
                return nodeRet
            else:

                if nodeRet['key'] == 'correlation_username' or nodeRet['key'] == 'target_username':
                    if 'NT AUTHORITY\\SYS' in value.upper():
                        value = value.replace('NT AUTHORITY\\SYSTEM', 'SYSTEM')
                    elif 'AUTHORI' in value.upper():
                        value = 'SYSTEM'

                nodeRet['args']['str']['value'] = value
                if notNode:
                    nodeRet["args"]["comparison"]["value"] = "!="
                else:
                    nodeRet['args']['comparison']['value'] = "="
                #return json.dumps(nodeRet)
                return nodeRet
        elif type(value) == list:
            return self.generateMapItemListNode(key, value, notNode)
        elif isinstance(value, SigmaTypeModifier):
            return self.generateMapItemTypedNode(key, value, notNode)
        elif value is None:
            #return self.nullExpression % (key, )
            #print("Performing null")
            #print(notNode)
            #print(key)
            nodeRet = { "key" : "empty", "description" : "Value Does Not Exist (IS NULL)", "class" : "function", "inputs" : { "comparison" : { "order" : 0, "source" : "comparison", "type" : "comparison" }, "column" : { "order" : 1, "source" : "columns", "type" : "str" } }, "args" : { "comparison" : { "value" : "!=" }, "column" : { "value" : "" } }, "return" : "boolean" }
            nodeRet['args']['column']['value'] = self.cleanKey(key).lower()
            nodeRet['description'] += " %s" % key
            if notNode:
                nodeRet['args']['comparison']['value'] = "!="
            else:
                nodeRet['args']['comparison']['value'] = "="
            #return json.dumps(nodeRet)
            #print(json.dumps(nodeRet))
            return nodeRet
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateMapItemListNode(self, key, value, notNode=False):
        if notNode:
            ret = { "id" : "and", "key": "And", "children" : [ ] }
        else:
            ret = { "id" : "or", "key": "Or", "children" : [ ] }
        for item in value:
            nodeRet = {"key": "",  "description": "", "class": "column", "return": "str", "args": { "comparison": { "value": "=" }, "str": { "value": "5" } } }
            nodeRet['key'] = self.cleanKey(key).lower()
            nodeRet['description'] = key
            nodeRet['rule_id'] = str(uuid.uuid4())
            if notNode:
                nodeRet['args']['comparison']['value'] = "!="
            else:
                nodeRet['args']['comparison']['value'] = "="
            if item is None:
                nodeRet['args']['str']['value'] = 'null'
                ret['children'].append( nodeRet )
            elif type(item) == str and "*" in item:
                if (nodeRet['key'] == 'correlation_username' or nodeRet['key'] == 'target_username'):
                    if 'NT AUTHORITY\\SYS' in item.upper():
                        item = item.replace('NT AUTHORITY\\SYSTEM', 'SYSTEM')
                    elif 'AUTHORI' in item.upper():
                        item = 'SYSTEM'
                item = item.replace("*", "EEEESTAREEE")
                item = re.escape(item)
                item = item.replace("EEEESTAREEE", ".*")
                endsWith = False
                startsWith = False
                if item[:2] == ".*":  
                    item = item[2:]
                    endsWith = True
                if item[-2:] == ".*":
                    item = item[:-2]
                    startsWith = True
                if item[-2:] == "\\\\":
                    item = item[:-2]


                if endsWith and not startsWith:
                    nodeRet['args']['str']['value'] = item + "$"
                elif startsWith and not endsWith:
                    nodeRet['args']['str']['value'] = "^" + item
                else:
                    nodeRet['args']['str']['value'] = item
                nodeRet['args']['str']['regex'] = "true"

                if notNode:
                    nodeRet["args"]["comparison"]["value"] = "!="
                else:
                    nodeRet['args']['comparison']['value'] = "="
                #print(item)
                ret['children'].append( nodeRet )
            else:
                nodeRet['args']['str']['value'] = self.generateValueNode(item, True)

                # custom, since we trim up  string size in log to save bytes
                key = nodeRet['key']
                value = nodeRet['args']['str']['value']

                if (nodeRet['key'] == 'correlation_username' or nodeRet['key'] == 'target_username'):
                    if 'NT AUTHORITY\\SYS' in value.upper():
                        value = value.replace('NT AUTHORITY\\SYSTEM', 'SYSTEM')
                    elif 'AUTHORI' in value.upper():
                        value = 'SYSTEM'

                if key == 'provider__name':
                    nodeRet['key'] = "product_name"
                    if value[0:17] == 'Microsoft-Windows':
                        value = value[18:]

                nodeRet['args']['str']['value'] = value

                ret['children'].append( nodeRet )
        retAnd = { "id" : "and", "key": "And", "children" : [ ret ] }
        return retAnd # '('+" or ".join(itemslist)+')'
        # return json.dumps(ret) # '('+" or ".join(itemslist)+')'

    def generateMapItemTypedNode(self, fieldname, value, notNode=False):
        nodeRet = {"key": "",  "description": "", "class": "column", "return": "str", "args": { "comparison": { "value": "=" }, "str": { "value": "5" } } }
        nodeRet['key'] = self.cleanKey(fieldname).lower()
        nodeRet['description'] = fieldname
        nodeRet['rule_id'] = str(uuid.uuid4())
        if type(value) == SigmaRegularExpressionModifier:
            value = self.generateValueNode(value, True)

            if (nodeRet['key'] == 'correlation_username' or nodeRet['key'] == 'target_username'):
                if 'NT AUTHORITY\\SYS' in value.upper():
                    value = value.replace('NT AUTHORITY\\SYSTEM', 'SYSTEM')
                elif 'AUTHORI' in value.upper():
                    value = 'SYSTEM'

            nodeRet['args']['str']['value'] = value
            nodeRet['args']['str']['regex'] = "true"
            if notNode:
                nodeRet["args"]["comparison"]["value"] = "!="
            else:
                nodeRet['args']['comparison']['value'] = "="
            return nodeRet
        else:
            raise NotImplementedError("Type modifier '{}' is not supported by backend".format(value.identifier))

    def generateValueNode(self, node, keypresent):
        return self.valueExpression % (self.cleanValue(str(node)))

    def generateNULLValueNode(self, node, notNode):
        # node.item
        nodeRet = { "key" : "empty", "description" : "Value Does Not Exist (IS NULL)", "class" : "function", "inputs" : { "comparison" : { "order" : 0, "source" : "comparison", "type" : "comparison" }, "column" : { "order" : 1, "source" : "columns", "type" : "str" } }, "args" : { "comparison" : { "value" : "!=" }, "column" : { "value" : node.item } }, "return" : "boolean" }
        nodeRet['args']['column']['value'] = self.cleanKey(node.item).lower()
        nodeRet['description'] += " %s" % key
        if notNode:
            nodeRet['args']['comparison']['value'] = "!="
        else:
            nodeRet['args']['comparison']['value'] = "="
        nodeRet['rule_id'] = str(uuid.uuid4())
        # return json.dumps(nodeRet)
        return nodeRet

    def generateNotNULLValueNode(self, node):
        # return self.notNullExpression % (node.item)
        return node.item

    def generateAggregation(self, agg, timeframe='00'):
        if agg == None:
            return None
        #print(agg.aggfunc)
        #print(type(agg.aggfunc))
        #print(agg.aggfunc_notrans)
        if not agg.aggfunc_notrans.lower() in ("count", "sum"):
            raise NotImplementedError("This aggregation operator '%s' has not been implemented" % agg.aggfunc_notrans)

        if agg.aggfunc == sigma.parser.condition.SigmaAggregationParser.AGGFUNC_NEAR:
            return None

        if agg.groupfield == None:
            agg.groupfield = "priority"

        if agg.groupfield != None and timeframe == '00':
            self.prefixAgg = " SELECT %s(%s) as agg_val from %s where " % (agg.aggfunc_notrans, self.cleanKey(agg.aggfield), self.aql_database)
            self.suffixAgg = " group by %s having agg_val %s %s" % (self.cleanKey(agg.groupfield), agg.cond_op, agg.condition)
            #print("Group field and timeframe is 00")
            min_count = 60
            nodeRet = {"key": "atomic_counter",  "description": self.cleanKey(agg.groupfield) + " %s aggregation stream counter" % agg.aggfunc_notrans, "class": "function", "return": "int",
                 "inputs": {
                     "columns" : { "order" : "0", "source" : "columns", "type" : "array", "objectKey" : "columns" },
                     "comparison" : { "order" : "1", "source" : "comparison", "type" : "comparison", "objectKey" : "comparison" },
                     "threshold" : { "order" : "2", "source" : "", "type" : "int", "objectKey" : "threshold" },
                     "limit" : { "order" : "3", "source" : "time_offset", "type" : "int", "objectKey" : "limit" },
                 },
                 "args": { 
                     "columns" : [ "ip_src", self.cleanKey(agg.groupfield) ],
                     "comparison": { "value": "%s" % agg.cond_op }, 
                     "threshold": { "value": int(agg.condition) }, 
                     "limit": { "value": min_count } 
                 } 
            }
            nodeRet['rule_id'] = str(uuid.uuid4())
            #print("No time range set")
            return nodeRet
        elif agg.groupfield != None and timeframe != None:
            for key, duration in self.generateTimeframe(timeframe).items():
                min_count = 60
                if key.lower() == 'hours':
                    min_count = 24 * int(duration)
                nodeRet = {"key": "atomic_counter",  "description": self.cleanKey(agg.groupfield) + " %s aggregation stream counter" % agg.aggfunc_notrans, "class": "function", "return": "int",
                         "inputs": {
                             "columns" : { "order" : "0", "source" : "columns", "type" : "array", "objectKey" : "columns" },
                             "comparison" : { "order" : "1", "source" : "comparison", "type" : "comparison", "objectKey" : "comparison" },
                             "threshold" : { "order" : "2", "source" : "", "type" : "int", "objectKey" : "threshold" },
                             "limit" : { "order" : "3", "source" : "time_offset", "type" : "int", "objectKey" : "limit" },
                         },
                         "args": { 
                             "columns" : [ "ip_src", self.cleanKey(agg.groupfield) ],
                             "comparison": { "value": "%s" % agg.cond_op }, 
                             "threshold": { "value": int(agg.condition) }, 
                             "limit": { "value": min_count } 
                         } 
                    }
                nodeRet['rule_id'] = str(uuid.uuid4())
                #self.prefixAgg = " SELECT %s(%s) as agg_val from %s where " % (agg.aggfunc_notrans, self.cleanKey(agg.aggfield), self.aql_database)
                #self.suffixAgg = " group by %s having agg_val %s %s LAST %s %s" % (self.cleanKey(agg.groupfield), agg.cond_op, agg.condition, duration, key)
                #print("Group field and timeframe")
                #return self.prefixAgg, self.suffixAgg
                return nodeRet
        else:
            self.prefixAgg = " SELECT %s(%s) as agg_val from %s where " % (agg.aggfunc_notrans, self.cleanKey(agg.aggfield), self.aql_database)
            self.suffixAgg = " group by %s having agg_val %s %s" % (self.cleanKey(agg.groupfield), agg.cond_op, agg.condition)
            #print("Last option")
            raise NotImplementedError("The 'agg' aggregation operator is not yet implemented for this backend") 
            return self.prefixAgg, self.suffixAgg
        #print(agg)
        raise NotImplementedError("The 'agg' aggregation operator is not yet implemented for this backend") 

    def generateTimeframe(self, timeframe):
        time_unit = timeframe[-1:]
        duration = timeframe[:-1]
        timeframe_object = {}
        if time_unit == "s":
            timeframe_object['seconds'] = int(duration)
        elif time_unit == "m":
            timeframe_object['minutes'] = int(duration)
        elif time_unit == "h":
            timeframe_object['hours'] = int(duration)
        elif time_unit == "d":
            timeframe_object['days'] = int(duration)
        else:
            timeframe_object['months'] = int(duration)
        return timeframe_object

    def generateBefore(self, parsed):
        if self.logname:
            return self.logname
        return self.logname

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        columns = list()
        mapped =None
        #print(sigmaparser.parsedyaml)
        self.logsource = sigmaparser.parsedyaml.get("logsource") if sigmaparser.parsedyaml.get("logsource") else sigmaparser.parsedyaml.get("logsources", {})
        fields = ""
        try:
            #print(sigmaparser.parsedyaml["fields"])
            for field in sigmaparser.parsedyaml["fields"]:
                mapped = sigmaparser.config.get_fieldmapping(field).resolve_fieldname(field, sigmaparser)
                if type(mapped) == str:
                    columns.append(mapped)
                elif type(mapped) == list:
                    columns.extend(mapped)
                else:
                    raise TypeError("Field mapping must return string or list")

            fields = ",".join(str(x) for x in columns)
            fields = " | table " + fields

        except KeyError:    # no 'fields' attribute
            mapped = None
            pass

        #print("Mapped: ", mapped)
        #print(sigmaparser.parsedyaml)
        #print(sigmaparser.condparsed)
        #print("Columns: ", columns)
        #print("Fields: ", fields)
        #print("Logsource: " , self.logsource)

        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed, sigmaparser)
            before = self.generateBefore(parsed)
            after = self.generateAfter(parsed)

            #print("Before: ", before)

            #print("Query: ", query)

            result = ""
            if before is not None:
                result = before
            if query is not None:
                result += query
            if after is not None:
                result += after

            return result

    def dedupeAnds(self, arr, parentAnd=False):
        # simple dedupe
        for i in range(0, len(arr)):
            if 'id' in arr[i] and arr[i]['id'].lower() == "and":
                arr[i]['children'] = self.dedupeAnds(arr[i]['children'])

                if len(arr[i]['children']) == 1 and 'id' in arr[i]['children'][0] and arr[i]['children'][0]['id'].lower() == "and":
                    arr[i] = arr[i]['children'][0]


        return arr

        """
        for i in range(0, len(arr)):
            if parentAnd and 'id' in arr[i] and arr[i]['id'].lower() == "and":
                isAnd = True
            else:
                isAnd = False
            
            if 'children' in arr[i]:
                arr[i]['children'] = self.dedupeAnds(arr['i']['children'], isAnd)

            if parentAnd and 'id' in arr[i] and arr[i]['id'].lower() == "and":
                pass

        if len(arr) == 1 and 'id' in arr[0] and arr[0]['id'].lower() == "and":
            # print("Returning less!")
            for i in range(0, len(arr) ):
                if 'id' in arr[i] and arr[i]['id'].lower() == "and":
                    arr[i]['children'] = self.dedupeAnds(arr[i]['children'])
            return arr[0]['children']

        """
        return arr

    """
    def dedupeAnds(self, arr, parentAnd=False):
        #if not parentAnd:
        #    for i in range(0, len(arr) ):
        #        if 'id' in arr[i] and arr[i]['id'].lower() == "and":
        #            arr[i]['children'] = self.dedupeAnds(arr[i]['children'], False)

        if len(arr) == 1 and 'id' in arr[0] and arr[0]['id'].lower() == "and":
            # print("Returning less!")
            for i in range(0, len(arr) ):
                if 'id' in arr[i] and arr[i]['id'].lower() == "and":
                    arr[i]['children'] = self.dedupeAnds(arr[i]['children'])
            return arr[0]['children']

        allAndCheck = True
        for i in range(0, len(arr) ):
            # print(arr[i])
            if 'id' in arr[i] and arr[i]['id'].lower() == "and":
                arr[i]['children'] = self.dedupeAnds(arr[i]['children'])
            else:
                allAndCheck = False


        x = [ ]
        if allAndCheck:
            for i in range(0, len(arr)):
                x = x + arr[i]['children']
            return x
        return arr
    """

    def generateQuery(self, parsed, sigmaparser):
        self.sigmaparser = sigmaparser
        result = self.generateNode(parsed.parsedSearch)
        prefix = ""
        ret = '[ { "id" : "and", "key": "And", "children" : ['
        ret2 = ' ] } ]'
        try:
            mappedFields = []
            for field in sigmaparser.parsedyaml["fields"]:
                    mapped = sigmaparser.config.get_fieldmapping(field).resolve_fieldname(field, sigmaparser)
                    #print(mapped)
                    mappedFields.append(mapped)
                    if " " in mapped and not "(" in mapped:
                        prefix += ", \"" + mapped + "\""
                    else:
                        prefix +=  ", " + mapped

        except KeyError:    # no 'fields' attribute
            mapped = None
            pass

        try:
            timeframe = sigmaparser.parsedyaml['detection']['timeframe']
        except:
            timeframe = None

        if parsed.parsedAgg and timeframe == None:
            addition = self.generateAggregation(parsed.parsedAgg)
            #print(addition)
            #print(result)
            if addition:
                if not 'children' in result:
                    rec = self.subExpression % json.dumps(result)
                    result = json.loads(rec)
                    #print(result)
                result['children'].append(addition)
            elif parsed.parsedAgg:
                #print(result)
                raise Exception("No agg returned, something is off")
        elif parsed.parsedAgg != None and timeframe != None:
            addition = self.generateAggregation(parsed.parsedAgg, timeframe)
            #print(addition)
            #print(result)
            if addition:
                #print(result)
                if not 'children' in result:
                    rec = self.subExpression % json.dumps(result)
                    result = json.loads(rec)
                    #print(result)
                result['children'].append(addition)
            elif parsed.parsedAgg:
                #print(result)
                raise Exception("No agg returned, something is off")
        else:
            # result = prefix + result
            pass

        result = json.dumps(result)

        analytic_txt = ret + result + ret2 # json.dumps(ret)
        try:
            analytic = json.loads(analytic_txt) # json.dumps(ret)
            # analytic = self.dedupeAnds(analytic)
            analytic[0]['children'] = self.dedupeAnds(analytic[0]['children'], True)

        except Exception as e:
            print("Failed to parse json: %s" % analytic_txt)
            raise Exception("Failed to parse json: %s" % analytic_txt)

        cmt = "Sigma Rule: %s\n" % sigmaparser.parsedyaml['id'] 
        if 'author' in sigmaparser.parsedyaml:
            cmt += "Author: %s\n" % sigmaparser.parsedyaml['author'] 
        else:
            cmt += "Author: Unknown\n"
        cmt += "Level: %s\n" % sigmaparser.parsedyaml['level'] 
        if 'falsepositives' in sigmaparser.parsedyaml and type(sigmaparser.parsedyaml['falsepositives']) is list:
            if len(sigmaparser.parsedyaml['falsepositives']) > 0:
                cmt += "False Positives: "
                for v in sigmaparser.parsedyaml['falsepositives']:
                    if v:
                        cmt += "%s, " % v
                    else:
                        cmt += "None, "
                cmt = cmt[:-2] + "\n"
        elif 'falsepositives' in sigmaparser.parsedyaml and sigmaparser.parsedyaml['falsepositives']:
            raise Exception("Unknown type for false positives: ", type(sigmaparser.parsedyaml['falsepositives']))

        if 'references' in sigmaparser.parsedyaml:
            ref = "%s" % "\n".join(sigmaparser.parsedyaml['references']) 
        else:
            ref = ''
        record = {
            "rules" : analytic, # analytic_txt.replace('"','""'),
            "filter_name" : sigmaparser.parsedyaml['title'],
            "filter_details" : cmt,
            "actions_category_name" : "Add (+)",
            "correlation_action" : 5.00,
            "date_added" : sigmaparser.parsedyaml['date'],
            "enabled" : False,
            # "enabled" : True,
            "public" : True,
            "references" : ref,
            "group_name" : ".",
            "tags" : [ "sigma" ],
            "hawk_id" : sigmaparser.parsedyaml['id']
        }
        if 'tags' in sigmaparser.parsedyaml and isinstance(sigmaparser.parsedyaml, list):
            mitre_tactics = [ item.replace("attack.", "") for item in sigmaparser.parsedyaml['tags'] ]
            if len(mitre_tactics) > 0:
                record["tags"] = record['tags'] + mitre_tactics
                # set 1st tactic and technique found

                mitre_tactics_filtered = [ ]
                for item in mitre_tactics:
                    if re.match("^t[0-9]+", item): mitre_tactics_filtered.append(item.upper())
                if len(mitre_tactics_filtered) > 0:
                    record["technique"] = mitre_tactics_filtered[0]
            
        score_reason_txt = "Scoring:\n"
        if not 'status' in self.sigmaparser.parsedyaml or 'status' in self.sigmaparser.parsedyaml and self.sigmaparser.parsedyaml['status'] != 'experimental':
            record['correlation_action'] += 5.0;
            score_reason_txt += "Status is not experimental (+5)\n"
        elif 'status' in self.sigmaparser.parsedyaml and self.sigmaparser.parsedyaml['status'] == 'experimental':
            record["tags"].append("qa")
            score_reason_txt += "Status is experimental (+0)\n"
        if 'falsepositives' in self.sigmaparser.parsedyaml and len(self.sigmaparser.parsedyaml['falsepositives']) > 1:
            record['correlation_action'] -= (2.0 * len(self.sigmaparser.parsedyaml['falsepositives']) )
            score_reason_txt += "False positives  (-2 * %r)\n" % len(self.sigmaparser.parsedyaml['falsepositives'])

        if 'level' in self.sigmaparser.parsedyaml:
            if self.sigmaparser.parsedyaml['level'].lower() == 'critical':
                record['correlation_action'] += 15.0;
                score_reason_txt += "Critical (+15)\n"
            elif self.sigmaparser.parsedyaml['level'].lower() == 'high':
                record['correlation_action'] += 10.0;
                score_reason_txt += "High (+10)\n"
            elif self.sigmaparser.parsedyaml['level'].lower() == 'medium' or self.sigmaparser.parsedyaml['level'].lower() == 'moderate':
                # record['correlation_action'] += 0.0;
                score_reason_txt += "Medium (+0)\n"
                pass
            elif self.sigmaparser.parsedyaml['level'].lower() == 'low':
                record['correlation_action'] -= 10.0;
                score_reason_txt += "Low (-10)\n"
            elif self.sigmaparser.parsedyaml['level'].lower() == 'informational':
                record['correlation_action'] -= 15.0;
                score_reason_txt += "Informational (-15)\n"

        record["filter_details"] += "\n\n" + score_reason_txt

        if record['correlation_action'] < 0.0:
            record['correlation_action'] = 0.0
       
        return json.dumps(record)

    def snake_case(self, str):
        res = [str[0].lower()]
        for c in str[1:]:
            if c in ('ABCDEFGHIJKLMNOPQRSTUVWXYZ'):
                res.append('_')
                res.append(c.lower())
            else:
                res.append(c)
         
        return ''.join(res)



# ee-outliers backend for sigmac
# NVISO (@NVISO_Labs)

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .elasticsearch import ElasticsearchDSLBackend
import json
import logging
import configparser
from .mixins import MultiRuleOutputMixin
from io import StringIO


class OutliersBackend(ElasticsearchDSLBackend, MultiRuleOutputMixin):
    """Converts Sigma rule into ee-outliers"""
    identifier = 'ee-outliers'
    active = True

    def generate(self, sigmaparser):
        super().generate(sigmaparser)

        self.tags = sigmaparser.parsedyaml.setdefault("tags", "")

        if len(self.queries) == 1:
            dsl = json.dumps(self.queries[0])
        else:
            dsl = json.dumps(self.queries)

        self.queries = []

        use_case_name = self.getRuleName(sigmaparser)

        index = ''
        if self.indices is not None and len(self.indices) == 1:
            index = self.indices[0]

        types = ["Sigma hit"]
        types.extend(self.tags)

        config_data = {
            "es_dsl_filter": dsl,
            "es_index": index,
            "outlier_type": ", ".join(types),
            "outlier_reason": "Sigma hit - " + self.title,
            "outlier_summary": "Sigma hit - " + self.title,
            "run_model": 1,
            "test_model": 0
        }

        config = configparser.ConfigParser(interpolation=None)
        config["simplequery_sigma_" + use_case_name] = config_data

        output = StringIO()
        config.write(output)
        result = output.getvalue()
        output.close()

        return result
    
    def finalize(self):
        """
        Is called after the last file was processed with generate(). The right place if this backend is not intended to
        look isolated at each rule, but generates an output which incorporates multiple rules, e.g. dashboards.
        """
        pass



# Output backends for sigmac
# Copyright 2022 Netmonastery, Inc.#

import re
from .base import SingleTextQueryBackend
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier
from sigma.parser.condition import SigmaAggregationParser


class DnifBackend(SingleTextQueryBackend):
    """Base class for DNIF backend"""
    identifier = "dnif"
    andToken = " and "
    orToken = " or "
    notToken = "not"
    subExpression = "%s"
    listExpression = "%s"
    listSeparator = " "
    valueExpression = "\"%s\""
    nullExpression = "NOT %s=\"*\""
    notNullExpression = "%s=\"*\""
    mapExpression = "%s == \"%s\""
    mapListsSpecialHandling = True
    mapListValueExpression = "%s IN %s"
    active = True

    config_required = True
    ymlFileName = None

    def __init__(self, sigmaconfig, options=None):
        """
        Initialize backend. This gets a sigmaconfig object, which is notified
        about the used backend class by
        passing the object instance to it.
        """
        super().__init__(sigmaconfig)
        self.table = None
        self.timeframe = None

    def generateANDNode(self, node):
        """
        Generates and nodes for query
        this method accepts the node and returns transformed node
        according to the query language
        """
        generated = [self.generateNode(val) for val in node]
        transformed = []
        for generated_node in generated:
            if generated_node is not None:
                if re.search(self.orToken, generated_node):
                    transformed.append("(" + generated_node + ")")
                else:
                    transformed.append(generated_node)
        return self.andToken.join(transformed)

    def default_value_mapping(self, val):
        """
        creates default value mapping for
        the rules. this method accepts any value
        and returns a transformed value
        """
        if isinstance(val, int):
            return f"== {val}"
        default_operator = "=="
        if isinstance(val, str) and val[1:-1]:
            if "*" in val[1:-1]:  # value contains * inside string - use regex match
                default_operator = ""
                val = re.sub(r'(\\\\\*|\*)', '.*', val)
                if "\\" in val:
                    val = f'@rlike("%", "{val}")'
                else:
                    val = f'rlike("%", "{val}")'
                return f'{default_operator} {self.cleanValue(val)}'
            elif val.startswith("*") or val.endswith("*"):
                default_operator = "like"
                if val.startswith("*") and val.endswith("*"):
                    val = f'%{val[1:-1]}%'
                elif val.startswith("*"):
                    val = f'%{val[1:]}'
                elif val.endswith("*"):
                    val = f'{val[:-1]}%'
                if "\\" in val:
                    return f'{default_operator} "{self.cleanValue(val)}"'
                return f'{default_operator} "{self.cleanValue(val)}"'
            elif "\\" in val:
                return f'{default_operator} @"{self.cleanValue(val)}"'
        elif isinstance(val, SigmaRegularExpressionModifier):
            default_operator = ""
            val = f'rlike("%", "{val}")'
            return f'{default_operator} {self.cleanValue(val)}'
        return f'{default_operator} "{self.cleanValue(val)}"'

    def generateORNode(self, node):
        """
        Generates or nodes for query
        this method accepts the node and returns transformed node
        according to the query language
        """
        generated = [self.generateNode(val) for val in node]
        transformed = {}
        transformed_query = []
        for generated_node in generated:
            if generated_node is not None:
                generated_node = generated_node.split(' == ')
                if len(generated_node) == 1:
                    transformed_query.append(generated_node[0])
                else:
                    if generated_node[0] not in transformed:
                        transformed[generated_node[0]] = [generated_node[1]]
                    else:
                        if generated_node[1] not in transformed[generated_node[0]]:
                            transformed[generated_node[0]].append(generated_node[1])
        if transformed:
            _transformed_query = [f'{key} IN ({", ".join(value)})'
                                  for key, value in transformed.items()]
            transformed_query.extend(_transformed_query)
        return self.orToken.join(transformed_query)

    def generateAggregation(self, agg):
        """
        Generates aggregations for query
        this method accepts the aggregation and
        returns a query with aggregation applied to it
        according to the query language
        """
        if agg is None:
            return ""
        if agg.aggfunc == SigmaAggregationParser.AGGFUNC_NEAR:
            raise NotImplementedError("The 'near' aggregation operator is not yet implemented"+
                                      "for this backend")

        if agg.groupfield is None:
            if agg.aggfunc_notrans == 'count':
                if agg.aggfield is None:
                    if agg.condition:
                        if self.timeframe:
                            return f" | select count(*) as count_col" \
                                   f" | having count_col {agg.cond_op} {agg.condition}" \
                                   f" | duration {self.timeframe}"
                        return f" | select count(*) as count_col" \
                               f" | having count_col {agg.cond_op} {agg.condition}"
                else:
                    if self.timeframe:
                        return f" | groupby {agg.groupfield}" \
                               f" | select {agg.groupfield}, count(*) as count_col" \
                               f" | having count_col {agg.cond_op} {agg.condition}" \
                               f" | duration {self.timeframe}"
                    return f" | groupby {agg.groupfield}" \
                           f" | select {agg.groupfield}, count(*) as count_col" \
                           f" | having count_col {agg.cond_op} {agg.condition}"
            if self.timeframe:
                return f' | groupby {agg.aggfield or ""}' \
                       f' | select {agg.aggfield or ""}, distinct_count({agg.aggfield or ""}), count(*) as total_count' \
                       f' | duration {self.timeframe}'

            return " | groupby %s" \
                   " | select %s, distinct_count(%s), count(*) " \
                   " as total_count" % (agg.aggfield or "",
                                        agg.aggfield or "",
                                        agg.aggfield or "")

        if agg.aggfunc_notrans == 'count':
            if agg.aggfield is None:
                if agg.condition:
                    if self.timeframe:
                        return " | groupby %s" \
                               " | select %s, count(*) as count_col" \
                               " | having count_col %s %s" \
                               " | duration %s" % (agg.groupfield,
                                                   agg.groupfield,
                                                   agg.cond_op,
                                                   agg.condition,
                                                   self.timeframe)
                    return " | groupby %s" \
                           " | select %s, count(*) as count_col" \
                           " | having count_col %s %s" % (agg.groupfield,
                                                          agg.groupfield,
                                                          agg.cond_op,
                                                          agg.condition)
            if self.timeframe:
                return " | groupby %s" \
                       " | select %s, count(%s)" \
                       " | duration %s" % (agg.groupfield or "",
                                           agg.groupfield or "",
                                           agg.aggfield or "",
                                           self.timeframe)
            return " | groupby %s" \
                   " | select %s, count(%s)" % (agg.groupfield or "",
                                                agg.groupfield or "",
                                                agg.aggfield or "")
        elif agg.aggfunc_notrans == 'sum':
            if agg.aggfield is None:
                if self.timeframe:
                    return " | groupby %s" \
                           " | select %s, sum(*) as count_col" \
                           " | having count_col %s %s" \
                           " | duration %s" % (agg.groupfield,
                                               agg.groupfield,
                                               agg.cond_op,
                                               agg.condition,
                                               self.timeframe)
                return " | groupby %s" \
                       " | select %s, sum(*) as count_col" \
                       " | having count_col %s %s" % (agg.groupfield,
                                                      agg.groupfield,
                                                      agg.cond_op,
                                                      agg.condition)
            else:
                if self.timeframe:
                    return " | groupby %s" \
                           " | select %s, sum(%s)" \
                           " | duration %s" % (agg.groupfield or "",
                                               agg.groupfield or "",
                                               agg.aggfield or "",
                                               self.timeframe)
                return " | groupby %s" \
                       " | select %s, sum(%s)" % (agg.groupfield or "",
                                                  agg.groupfield or "",
                                                  agg.aggfield or "")

    def generateMapItemNode(self, node):
        key, value = node
        key = self.fieldNameMapping(key, value)
        # handle map items with values list like multiple OR-chained conditions
        if type(value) == list:
            return self.generateORNode(
                    [(key, v) for v in value]
                    )
        elif type(value) in (str, int) or isinstance(value, SigmaRegularExpressionModifier):    # default value processing'
            value_mapping = self.default_value_mapping
            mapping = (key, value_mapping)
            if len(mapping) == 1:
                mapping = mapping[0]
                if type(mapping) == str:
                    return mapping
                elif callable(mapping):
                    return self.generateSubexpressionNode(
                            self.generateANDNode(
                                [cond for cond in mapping(key, self.cleanValue(value))]
                                )
                            )
            elif len(mapping) == 2:
                result = list()
                # iterate mapping and mapping source value synchronously over key and value
                for mapitem, val in zip(mapping, node):
                    if type(mapitem) == str:
                        result.append(mapitem)
                    elif callable(mapitem):
                        mapitem_value = mapitem(self.cleanValue(val))
                        if 'rlike' in mapitem_value:
                            mapitem_value = re.sub(r'\"%\"', result[0], mapitem_value)
                        result.append(mapitem_value)
                for res in result:
                    if 'rlike' in res:
                        result[0] = ''
                return "{} {}".format(*result)
            else:
                raise TypeError("Backend does not support map values of type " + str(type(value)))
        else:
            return super().generateMapItemNode(node)

    def generateNOTNode(self, node):
        generated = self.generateNode(node.item)
        if generated is not None:
            return "%s %s" % (self.notToken, generated)
        else:
            return None

    def generateMapItemListNode(self, key, value):
        if isinstance(value, SigmaRegularExpressionModifier):
            key_mapped = self.fieldNameMapping(key, value)
            return {'regexp': {key_mapped: str(value)}}
        if not set([type(val) for val in value]).issubset({str, int}):
            raise TypeError("List values must be strings or numbers")
        if isinstance(value, list):
            if 'or' in value:
                self.generateORNode(value)
            elif 'and' in value:
                self.generateANDNode(value)
        return ' or '.join(['%s=%s' % (key, self.generateValueNode(item)) for item in value])

    def generateTypedValueNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateNULLValueNode(self, fieldname):
        return self.nullExpression % fieldname

    def generateNotNULLValueNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateBefore(self, parsed):
        return "stream=%s where " % self.table

    def getTable(self, parsed_rule_data):
        logsource_data = parsed_rule_data.get('logsource')
        if logsource_data.get('category'):
            self.table = logsource_data.get('category')
        elif logsource_data.get('product'):
            self.table = logsource_data.get('product')
        elif logsource_data.get('service'):
            self.table = logsource_data.get('service')

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        parsed_yaml = sigmaparser.parsedyaml
        if parsed_yaml.get('detection').get('timeframe'):
            self.timeframe = parsed_yaml['detection']['timeframe']

        if sigmaparser.get_logsource() and sigmaparser.get_logsource().index:
            self.table = sigmaparser.get_logsource().index[0]
        else:
            self.getTable(parsed_yaml)
        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed)
            before = self.generateBefore(parsed)

            result = ""
            if before is not None:
                result = before
            if query is not None:
                result += query
            if result.endswith(" | "):
                result = result.strip(" | ")
            return result



# Output backends for sigmac
# Copyright 2018 Thomas Patzke

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import sigma
from functools import wraps
from .base import SingleTextQueryBackend
from .exceptions import NotSupportedError
from ..parser.modifiers.base import SigmaTypeModifier


def wrapper(method):
    @wraps(method)
    def _impl(self, method_args):
        key, value, *_ = method_args
        if '.keyword' in key:
            key = key.split('.keyword')[0]
        if key not in self.skip_fields:
            method_output = method(self, method_args)
            return method_output
        else:
            return
    return _impl

class WindowsDefenderATPBackend(SingleTextQueryBackend):
    """Converts Sigma rule into Microsoft Defender ATP Hunting Queries."""
    identifier = "mdatp"
    active = True
    config_required = False

    reEscape = re.compile('(?:\\\\)?(")')
    reClear = None
    andToken = " and "
    orToken = " or "
    notToken = "not "
    subExpression = "(%s)"
    listExpression = "(%s)"
    listSeparator = ", "
    valueExpression = "\"%s\""
    nullExpression = "isnull(%s)"
    notNullExpression = "isnotnull(%s)"
    mapExpression = "%s == %s"
    mapListsSpecialHandling = True
    mapListValueExpression = "%s in %s"

    skip_fields = {
        "Description",
        "_exists_",
        "FileVersion",
        "Product",
        "Company",
        "IMPHASH",
    }

    def __init__(self, *args, **kwargs):
        """Initialize field mappings"""
        super().__init__(*args, **kwargs)
        self.fieldMappings = {       # mapping between Sigma and ATP field names
            # Supported values:
            # (field name mapping, value mapping): distinct mappings for field name and value, may be a string (direct mapping) or function maps name/value to ATP target value
            # (mapping function,): receives field name and value as parameter, return list of 2 element tuples (destination field name and value)
            # (replacement, ): Replaces field occurrence with static string
            "DeviceProcessEvents": {
                "AccountName": (self.id_mapping, self.default_value_mapping),
                "CommandLine": ("ProcessCommandLine", self.default_value_mapping),
                "DeviceName": (self.id_mapping, self.default_value_mapping),
                "EventType": ("ActionType", self.default_value_mapping),
                "FileName": (self.id_mapping, self.default_value_mapping),
                "Image": ("FolderPath", self.default_value_mapping),
                "ImagePath": ("FolderPath", self.default_value_mapping),
                "ImageLoaded": ("FolderPath", self.default_value_mapping),
                "LogonType": (self.id_mapping, self.logontype_mapping),
                "NewProcessName": ("FolderPath", self.default_value_mapping),
                "OriginalFileName": ("ProcessVersionInfoOriginalFileName", self.default_value_mapping),
                "ParentCommandLine": ("InitiatingProcessCommandLine", self.default_value_mapping),
                "ParentName": ("InitiatingProcessFileName", self.default_value_mapping),
                "ParentProcessName": ("InitiatingProcessFileName", self.default_value_mapping),
                "ParentImage": ("InitiatingProcessFolderPath", self.default_value_mapping),
                "IntegrityLevel": ("ProcessIntegrityLevel", self.default_value_mapping),
                "SourceImage": ("InitiatingProcessFolderPath", self.default_value_mapping),
                "TargetImage": ("FolderPath", self.default_value_mapping),
                "User": (self.decompose_user, ),
            },
            "DeviceEvents": {
                "AccountName": (self.id_mapping, self.default_value_mapping),
                "CommandLine": ("ProcessCommandLine", self.default_value_mapping),
                "DestinationHostname":  ("RemoteUrl", self.default_value_mapping),
                "DestinationIp": ("RemoteIP", self.default_value_mapping),
                "DestinationPort": ("RemotePort", self.porttype_mapping),
                "EventType": ("ActionType", self.default_value_mapping),
                "FileName": (self.id_mapping, self.default_value_mapping),
                "ParentCommandLine": ("InitiatingProcessCommandLine", self.default_value_mapping),
                "ParentProcessName": ("InitiatingProcessParentFileName", self.default_value_mapping),
                "ProcessName": ("InitiatingProcessFileName", self.default_value_mapping),
                "ServiceFileName": ("FileName", self.default_value_mapping),
                "SourceIp": ("LocalIP", self.default_value_mapping),
                "SourcePort": ("LocalPort", self.porttype_mapping),
                "TargetFilename": ("FolderPath", self.default_value_mapping),
                "TargetObject": ("RegistryKey", self.default_value_mapping),
                "TargetImage": ("FolderPath", self.default_value_mapping),
                "Image": ("InitiatingProcessFolderPath", self.default_value_mapping),
                "User":  (self.decompose_user, ),
            },
            "DeviceRegistryEvents": {
                "DataType": ("RegistryValueType", self.default_value_mapping),
                "Details": ("RegistryValueData", self.default_value_mapping),
                "EventType": ("ActionType", self.default_value_mapping),
                "Image": ("InitiatingProcessFolderPath", self.default_value_mapping),
                "CommandLine": ("InitiatingProcessCommandLine", self.default_value_mapping),
                "ObjectValueName": ("RegistryValueName", self.default_value_mapping),
                "ProcessName": ("InitiatingProcessFileName", self.default_value_mapping),
                "ParentName": ("InitiatingProcessParentFileName", self.default_value_mapping),
                "ParentProcessName": ("InitiatingProcessParentFileName", self.default_value_mapping),
                "TargetObject": ("RegistryKey", self.default_value_mapping),
                "User":  (self.decompose_user, ),
            },
            "DeviceFileEvents": {
                "FileName": (self.id_mapping, self.default_value_mapping),
                "OriginIp": ("FileOriginIp", self.default_value_mapping),
                "OriginReferrer": ("FileOriginReferrerUrl", self.default_value_mapping),
                "OriginUrl": ("FileOriginUrl", self.default_value_mapping),
                "Image": ("InitiatingProcessFolderPath", self.default_value_mapping),
                "CommandLine": ("InitiatingProcessCommandLine", self.default_value_mapping),
                "ProcessName": ("InitiatingProcessFileName", self.default_value_mapping),
                "ParentName": ("InitiatingProcessParentFileName", self.default_value_mapping),
                "ParentProcessName": ("InitiatingProcessParentFileName", self.default_value_mapping),
                "TargetFilename": ("FolderPath", self.default_value_mapping),
                "User":  (self.decompose_user, ),
            },
            "DeviceNetworkEvents": {
                "DestinationHostname": ("RemoteUrl", self.default_value_mapping),
                "DestinationIp": ("RemoteIP", self.default_value_mapping),
                "DestinationIsIpv6": ("RemoteIP has \":\"", ),
                "DestinationPort": ("RemotePort", self.porttype_mapping),
                "DeviceName": (self.id_mapping, self.default_value_mapping),
                "EventType": ("ActionType", self.default_value_mapping),
                "Image": ("InitiatingProcessFolderPath", self.default_value_mapping),
                "CommandLine": ("InitiatingProcessCommandLine", self.default_value_mapping),
                "Initiated": ("RemotePort", self.default_value_mapping),
                "ProcessName": ("InitiatingProcessFileName", self.default_value_mapping),
                "Protocol": ("RemoteProtocol", self.default_value_mapping),
                "SourceIp": ("LocalIP", self.default_value_mapping),
                "SourcePort": ("LocalPort", self.porttype_mapping),
                "User":  (self.decompose_user, ),
            },
            "DeviceImageLoadEvents": {
                "DeviceName": (self.id_mapping, self.default_value_mapping),
                "EventType": ("ActionType", self.default_value_mapping),
                "FileName": (self.id_mapping, self.default_value_mapping),
                "Image": ("InitiatingProcessFolderPath", self.default_value_mapping),
                "ImageLoaded": ("FolderPath", self.default_value_mapping),
                "CommandLine": ("InitiatingProcessCommandLine", self.default_value_mapping),
                "ParentProcessName": ("InitiatingProcessParentFileName", self.default_value_mapping),
                "ProcessName": ("InitiatingProcessFileName", self.default_value_mapping),
                "TargetImage": ("FolderPath", self.default_value_mapping),
                "User":  (self.decompose_user, ),
            }
        }
        self.current_table = ""

    def generateANDNode(self, node):
        generated = [ self.generateNode(val) for val in node ]
        filtered = []
        for g in generated:
            if g and g.startswith("ActionType"):
                if not any([i for i in filtered if i.startswith("ActionType")]):
                    filtered.append(g)
                else:
                    continue
            elif g is not None:
                filtered.append(g)
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            return self.andToken.join(filtered)
        else:
            return None

    def id_mapping(self, src):
        """Identity mapping, source == target field name"""
        return src

    def default_value_mapping(self, val):
        op = "=~"
        if type(val) == str:
            if "*" in val[1:-1]:
                # value contains * inside string - use regex match
                op = "matches regex"
                val = re.sub('([".^$]|\\\\(?![*?]))', '\\\\\g<1>', val)
                val = re.sub('\\*', '.*', val)
                val = re.sub('\\?', '.', val)
            else:
                # value possibly only starts and/or ends with *, use prefix/postfix match
                if val.endswith("*") and val.startswith("*"):
                    op = "contains"
                    val = self.cleanValue(val[1:-1])
                elif val.endswith("*"):
                    op = "startswith"
                    val = self.cleanValue(val[:-1])
                elif val.startswith("*"):
                    op = "endswith"
                    val = self.cleanValue(val[1:])

        if "\\" in val:
            val = val.replace("\\\\","\\")
            return "%s @\"%s\"" % (op, val)

        return "%s \"%s\"" % (op, val)

    def porttype_mapping(self, val):
        return "%s \"%s\"" % ("==", val)

    def logontype_mapping(self, src):
        """Value mapping for logon events to reduced ATP LogonType set"""
        logontype_mapping = {
            2: "Interactive",
            3: "Network",
            4: "Batch",
            5: "Service",
            7: "Interactive",   # unsure
            8: "Network",
            9: "Interactive",   # unsure
            10: "Remote interactive (RDP) logons",  # really the value?
            11: "Interactive"
        }
        try:
            return logontype_mapping[int(src)]
        except KeyError:
            raise NotSupportedError("Logon type %d unknown and can't be mapped" % src)

    def decompose_user(self, src_field, src_value):
        """Decompose domain\\user User field of Sysmon events into ATP InitiatingProcessAccountDomain and InititatingProcessAccountName."""
        reUser = re.compile("^(.*?)\\\\(.*)$")
        m = reUser.match(src_value)
        if m:
            domain, user = m.groups()
            return (("InitiatingProcessAccountDomain",  self.default_value_mapping(domain)), ("InititatingProcessAccountName",  self.default_value_mapping(user)))
        else:   # assume only user name is given if backslash is missing
            return (("InititatingProcessAccountName", self.default_value_mapping(src_value)))

    def generate(self, sigmaparser):
        self.tables = []
        try:
            self.category = sigmaparser.parsedyaml['logsource'].setdefault('category', None)
            self.product = sigmaparser.parsedyaml['logsource'].setdefault('product', None)
            self.service = sigmaparser.parsedyaml['logsource'].setdefault('service', None)
        except KeyError:
            self.category = None
            self.product = None
            self.service = None

        if (self.category, self.service) == ("process_creation", None) and self.product in ['windows', 'linux', 'macos']:
            self.tables.append("DeviceProcessEvents")
            self.current_table = "DeviceProcessEvents"
        elif (self.category, self.product, self.service) == ("registry_event", "windows", None):
            self.tables.append("DeviceRegistryEvents")
            self.current_table = "DeviceRegistryEvents"
        elif (self.category, self.product, self.service) == ("registry_add", "windows", None):
            self.tables.append("DeviceRegistryEvents")
            self.current_table = "DeviceRegistryEvents"
        elif (self.category, self.product, self.service) == ("registry_delete", "windows", None):
            self.tables.append("DeviceRegistryEvents")
            self.current_table = "DeviceRegistryEvents"
        elif (self.category, self.product, self.service) == ("registry_set", "windows", None):
            self.tables.append("DeviceRegistryEvents")
            self.current_table = "DeviceRegistryEvents"
        elif (self.category, self.product, self.service) == ("registry_rename", "windows", None):
            self.tables.append("DeviceRegistryEvents")
            self.current_table = "DeviceRegistryEvents"    
        elif (self.category, self.service) == ("file_event", None) and self.product in ['windows', 'linux', 'macos']:
            self.tables.append("DeviceFileEvents")
            self.current_table = "DeviceFileEvents"
        elif (self.category, self.product, self.service) == ("image_load", "windows", None):
            self.tables.append("DeviceImageLoadEvents")
            self.current_table = "DeviceImageLoadEvents"
        elif (self.category, self.service) == ("network_connection", None) and self.product in ['windows', 'linux', 'macos']:
            self.tables.append("DeviceNetworkEvents")
            self.current_table = "DeviceNetworkEvents"
        elif (self.category, self.product, self.service) == (None, "windows", "powershell"):
            self.tables.append("DeviceEvents")
            self.current_table = "DeviceEvents"
            self.orToken = ", "
        elif (self.category, self.product, self.service) == (None, "windows", "security"):
            self.tables.append("DeviceAlertEvents")
            self.current_table = "DeviceAlertEvents"

        return super().generate(sigmaparser)

    def generateBefore(self, parsed):
        if not any(self.tables):
            raise NotSupportedError("No MDATP table could be determined from Sigma rule")
        # if self.tables in "DeviceEvents" and self.service == "powershell":
        #     return "%s | where tostring(extractjson('$.Command', AdditionalFields)) in~ " % self.tables
        if len(self.tables) == 1:
            if self.tables[0] == "DeviceEvents" and self.service == "powershell":
                return "%s | where tostring(extractjson('$.Command', AdditionalFields)) in~ " % self.tables
            return "%s | where " % self.tables[0]
        else:
            if "DeviceEvents" in self.tables and self.service == "powershell":
                return "union %s | where tostring(extractjson('$.Command', AdditionalFields)) in~ " % ", ".join(self.tables)
            return "union %s | where " % ", ".join(self.tables)

    def generateORNode(self, node):
        generated = super().generateORNode(node)
        if generated:
            return "%s" % generated
        return generated

    def cleanValue(self, val):
        if self.reEscape:
            val = self.reEscape.sub(self.escapeSubst, val)
        return val

    def mapEventId(self, event_id):
        if self.product == "windows":
            if self.service == "sysmon" and event_id == 1 \
                    or self.service == "security" and event_id == 4688:  # Process Execution
                self.tables.append("DeviceProcessEvents")
                self.current_table = "DeviceProcessEvents"
                return None
            elif self.service == "sysmon" and event_id == 3:  # Network Connection
                self.tables.append("DeviceNetworkEvents")
                self.current_table = "DeviceNetworkEvents"
                return None
            elif self.service == "sysmon" and event_id == 7:  # Image Load
                self.tables.append("DeviceImageLoadEvents")
                self.current_table = "DeviceImageLoadEvents"
                return None
            elif self.service == "sysmon" and event_id == 8:  # Create Remote Thread
                self.tables.append("DeviceEvents")
                self.current_table = "DeviceEvents"
                return "ActionType == \"CreateRemoteThreadApiCall\""
            elif self.service == "sysmon" and event_id == 11:  # File Creation
                self.tables.append("DeviceFileEvents")
                self.current_table = "DeviceFileEvents"
                return "ActionType == \"FileCreated\""
            elif self.service == "sysmon" and event_id == 23:  # File Deletion
                self.tables.append("DeviceFileEvents")
                self.current_table = "DeviceFileEvents"
                return "ActionType == \"FileDeleted\""
            elif self.service == "sysmon" and event_id == 12:  # Create/Delete Registry Value
                self.tables.append("DeviceRegistryEvents")
                self.current_table = "DeviceRegistryEvents"
                return None
            elif self.service == "sysmon" and event_id == 13 \
                    or self.service == "security" and event_id == 4657:  # Set Registry Value
                self.tables.append("DeviceRegistryEvents")
                self.current_table = "DeviceRegistryEvents"
                return "ActionType == \"RegistryValueSet\""
            elif self.service == "security" and event_id == 4624:
                self.tables.append("DeviceLogonEvents")
                self.current_table = "DeviceLogonEvents"
                return None
            elif self.service == "system" and event_id == 7045: # New Service Install
                self.tables.append("DeviceEvents")
                self.current_table = "DeviceEvents"
                return "ActionType == \"ServiceInstalled\""
            else:
                if not self.tables:
                    raise NotSupportedError("No sysmon Event ID provided")
                else:
                    raise NotSupportedError("No mapping for Event ID %s" % event_id)

    @wrapper
    def generateMapItemNode(self, node):
        """
        ATP queries refer to event tables instead of Windows logging event identifiers. This method catches conditions that refer to this field
        and creates an appropriate table reference.
        """
        key, value = node
        if key == "EventID":
            # EventIDs are not reflected in condition but in table selection
            if isinstance(value, str) or isinstance(value, int):
                value = int(value) if isinstance(value, str) else value
                return self.mapEventId(value)
            elif isinstance(value, list):
                return_payload = []
                for event_id in value:
                    res = self.mapEventId(event_id)
                    if res:
                        return_payload.append(res)
                if len(return_payload) == 1:
                    return return_payload[0]
                elif not any(return_payload):
                    return None
                else:
                    return "(%s)" % self.generateORNode(
                    [(key, v) for v in value]
                    )
        if type(value) == list:         # handle map items with values list like multiple OR-chained conditions
            return "(%s)" % self.generateORNode(
                    [(key, self.cleanValue(v)) for v in value]
                    )
        elif type(value) in (str, int):     # default value processing
            try:
                mapping = self.fieldMappings[self.current_table][key]
            except KeyError:
                raise NotSupportedError("No mapping defined for field '%s' in '%s'" % (key, self.tables))
            if len(mapping) == 1:
                mapping = mapping[0]
                if type(mapping) == str:
                    return mapping
                elif callable(mapping):
                    conds = mapping(key, self.cleanValue(value))
                    return self.andToken.join(["{} {}".format(*cond) for cond in conds])
            elif len(mapping) == 2:
                result = list()
                for mapitem, val in zip(mapping, node):     # iterate mapping and mapping source value synchronously over key and value
                    if type(mapitem) == str:
                        result.append(mapitem)
                    elif callable(mapitem):
                        result.append(mapitem(self.cleanValue(val)))
                return "{} {}".format(*result)
            else:
                raise TypeError("Backend does not support map values of type " + str(type(value)))
        elif isinstance(value, SigmaTypeModifier):
            try:
                mapping = self.fieldMappings[self.current_table][key]
            except KeyError:
                raise NotSupportedError("No mapping defined for field '%s' in '%s'" % (key, self.tables))
            return self.generateMapItemTypedNode(mapping[0], value)

        return super().generateMapItemNode(node)

    def generateAggregation(self, agg):
        if agg == None:
            return ""
        if agg.aggfunc == sigma.parser.condition.SigmaAggregationParser.AGGFUNC_NEAR:
            raise NotImplementedError("The 'near' aggregation operator is not yet implemented for this backend")
        if agg.groupfield == None:
            if agg.aggfunc_notrans == 'count':
                if agg.aggfield == None :
                    return " | summarize val=count() | where val %s %s" % (agg.cond_op, agg.condition)
                else:
                    agg.aggfunc_notrans = 'dcount'
            return " | summarize val=%s(%s) as val | where val %s %s" % (agg.aggfunc_notrans, agg.aggfield or "", agg.cond_op, agg.condition)
        else:
            if agg.aggfunc_notrans == 'count':
                if agg.aggfield == None :
                    return " | summarize val=count() by %s | where val %s %s" % (agg.groupfield, agg.cond_op, agg.condition)
                else:
                    agg.aggfunc_notrans = 'dcount'
            return " | summarize val=%s(%s) by %s | where val %s %s" % (agg.aggfunc_notrans, agg.aggfield or "", agg.groupfield or "", agg.cond_op, agg.condition)



# Discovery of modifier modules
# Copyright 2016-2019 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from sigma.tools import getAllSubclasses, getClassDict
from .base import SigmaModifier

def getModifierList():
    """Return list of modifier classes"""
    path = os.path.dirname(__file__)
    return getAllSubclasses(path, "parser.modifiers", SigmaModifier)

modifiers = getClassDict(getModifierList())

def apply_modifiers(value, modifier_list):
    """
    Apply modifiers to value.

    value: value from Sigma rule
    modifiers: list of modifier names
    """
    for modifier in modifier_list:      # apply modifiers in given order
        value = modifiers[modifier](value).apply()
    return value



# Sigma parser
# Copyright 2016-2018 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .exceptions import SigmaParseError

class SimpleParser:
    """
    Rule-defined parser that converts a token stream into a Python object.

    Rules are defined in the class property parsingrules, a list of dict of tuples with the following format:
    [ { token_0_0: parsing_rule_0_0, token_0_1: parsing_rule_0_1, ..., token_0_n: parsing_rule_0_n } , ... , { token_m_0: parsing_rule_m_0, ... } ]

    Each list index of parsing rules represents a parser state.
    Each parser state is defined by a dict with associates a token with a rule definition.
    The rule definition is a tuple that defines what is done next when the parser encounters a token in the current parser state:

    ( storage attribute, transformation function, next ruleset)

    * storage attribute: the name of the object attribute that is used for storage of the attribute
    * transformation method: name of an object method that is called before storage. It gets a parameter and returns the value that is stored
    * next state: next parser state

    A None value means that the action (transformation, storage or state change) is not conducted.

    A negative state has the special meaning that no further token is expected and may be used as return value.
    The set or list finalstates contains valid final states. The parser verifies after the last token that it
    has reached one of these states. if not, a parse error is raised.
    """

    def __init__(self, tokens, init_state=0):
        self.state = init_state

        for token in tokens:
            if self.state < 0:
                raise SigmaParseError("No further token expected, but read %s" % (str(token)))
            try:
                rule = self.parsingrules[self.state][token.type]
            except KeyError as e:
                raise SigmaParseError("Unexpected token %s at %d in aggregation expression" % (str(token), token.pos)) from e

            value = token.matched
            trans_value = value
            if rule[1] != None:
                trans_value = getattr(self, rule[1])(value)
            if rule[0] != None:
                setattr(self, rule[0], trans_value)
                setattr(self, rule[0] + "_notrans", value)
            if rule[2] != None:
                self.state = rule[2]
        if self.state not in self.finalstates:
            raise SigmaParseError("Unexpected end of aggregation expression, state=%d" % (self.state))

    def __str__(self):  # pragma: no cover
        return "[ Parsed: %s ]" % (" ".join(["%s=%s" % (key, val) for key, val in self.__dict__.items() ]))



# Sigma value modifiers
# Copyright 2019 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum, auto
from .exceptions import SigmaModifierValueTypeError

class SigmaModifierTypes(Enum):
    NONE = auto()
    TRANSFORM = auto()
    TYPE = auto()

class SigmaModifier(object):
    """
    Value modifier base class. There are two modifier types (with separate base classes):

    identifier: identifier string that is used in Sigma rules to use modifier
    active: boolean if modifier will be found by autodiscovery (and therefore usable)
    modifier_type: type from SigmaModifierTypes enumeration
        * Value transformation modifiers: the value or list of values is transformed to a different
            value or a list of values. Base class: SigmaValueModifier
        * Type modifiers: modify the type of the modifier, e.g. to support regular expressions. The
            backend must handle these modifiers accordingly. Base class: SigmaTypeModifier
    valid_input_types: list of valid input types. Can be expected Python type (like str, int) or
        modifier class. object = don't care about type.
    """
    identifier = "base"
    active = False
    modifier_type = SigmaModifierTypes.NONE
    valid_input_types = (object,)

    def __init__(self, value):
        """Initialize modifier class. Store value or result of value transformation."""
        self.value = value
        if not self.validate():
            raise SigmaModifierValueTypeError

    def validate(self):
        """Validate if modifier is applicable to value. Expects that value is stored in self.value."""
        return any(( self.isinstance_recursive(self.value, t) for t in self.valid_input_types ))

    def isinstance_recursive(self, value: object, value_type: type) -> bool:
        """Validate values inside of a list or a tuple"""
        if not isinstance(value, value_type):
            return False

        if isinstance(value, (list, tuple, )):
            for val in value:
                if not any(( self.isinstance_recursive(val, t) for t in self.valid_input_types)):
                    return False

        return True

    def apply(self):
        """
        Apply modifier to value. This method can:
        * Return a transformed value
        * Return a list of values
        * Return an object from sigma.parser.condition that replaces the value in the condition tree
        * Return a value that is used by dedicated type modifier code in the backend to get a final
            query statement (only for type modifiers)

        The base method simply returns the stored value.
        """
        try:
            return self.value.apply()
        except AttributeError:
            return self.value

class SigmaTransformModifier(SigmaModifier):
    """Transform a value into a different value or a list of values"""
    identifier = "tansform_base"
    modifier_type = SigmaModifierTypes.TRANSFORM

class SigmaTypeModifier(SigmaModifier):
    """Modify the type of the value. This must be handled by the backend."""
    identifier = "type_base"
    modifier_type = SigmaModifierTypes.TYPE

    def apply(self):
        return self

    def __str__(self):
        return str(self.value)



import json
import re
import sigma
from sigma.backends.base import SingleTextQueryBackend
from sigma.parser.condition import SigmaAggregationParser, NodeSubexpression, ConditionAND, ConditionOR, ConditionNOT
from sigma.parser.exceptions import SigmaParseError
from .mixins import MultiRuleOutputMixin
from sigma.parser.modifiers.transform import SigmaContainsModifier, SigmaStartswithModifier, SigmaEndswithModifier
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier
from ..parser.modifiers.base import SigmaTypeModifier

gUnsupportedCategories = {}


def convert_sigma_level_to_uberagent_risk_score(level):
    """Converts the given Sigma rule level to uberAgent ESA RiskScore property."""
    levels = {
        "critical": 100,
        "high": 75,
        "medium": 50,
        "low": 25
    }

    if level in levels:
        return levels[level]

    return 0


def convert_sigma_name_to_uberagent_tag(name):
    """Converts the given Sigma rule name to uberAgent ESA Tag property."""
    tag = name.lower().replace(" ", "-")
    tag = re.sub(r"-{2,}", "-", tag, 0, re.IGNORECASE)
    return tag


def convert_sigma_category_to_uberagent_event_type(category):
    categories = {
        "process_creation": "Process.Start",
        "image_load": "Image.Load",
        "dns": "Dns.Query",
        "dns_query": "Dns.Query",
        "network_connection": "Net.Any",
        "firewall": "Net.Any",
        "create_remote_thread": "Process.CreateRemoteThread",
        "registry_event": "Reg.Any",
        "registry_add": "Reg.Any",
        "registry_delete": "Reg.Any",
        "registry_set": "Reg.Any",
        "registry_rename": "Reg.Any"
    }

    if category in categories:
        return categories[category]

    if category in gUnsupportedCategories:
        gUnsupportedCategories[category] += 1
    else:
        gUnsupportedCategories[category] = 1

    return None


def is_sigma_category_supported(category):
    """Returns whether uberAgent ESA knows the given category or not."""
    return convert_sigma_category_to_uberagent_event_type(category) is not None


class IgnoreTypedModifierException(Exception):
    """
    IgnoreTypedModifierException
    Helper class to ignore exceptions of type identifiers that are not yet supported.
    """
    pass


class IgnoreFieldException(Exception):
    """
    IgnoreFieldException
    Helper class to ignore exceptions of specific fields that are not yet supported.
    """
    pass


class IgnoreAggregationException(Exception):
    """
    IgnoreAggregationException
    Helper class to ignore exceptions of aggregation rules that are not yet supported.
    """


class MalformedRuleException(Exception):
    """
    MalformedRuleException
    Helper class to ignore exceptions of malformed rules.
    """
    pass


class ActivityMonitoringRule:
    """
    ActivityMonitoringRule
    This class wraps a [ActivityMonitoringRule] configuration block.
    """

    def __init__(self):
        self.id = ""
        self.name = ""
        self.event_type = None
        self.tag = ""
        self.query = ""
        self.risk_score = 0
        self.description = ""
        self.sigma_level = ""
        self.annotation = ""
        self.generic_properties = []

    def set_id(self, id):
        """Sets the RuleId property."""
        self.id = id

    def set_query(self, query):
        """Sets the generated query property."""
        self.query = query

    def set_name(self, name):
        """Sets the RuleName."""
        self.name = name

    def set_tag(self, tag):
        """Sets the Tag property."""
        self.tag = tag

    def set_event_type(self, event_type):
        """Sets the EventType property."""
        self.event_type = event_type

    def set_risk_score(self, risk_score):
        """Sets the RiskScore property."""
        self.risk_score = risk_score

    def set_sigma_level(self, level):
        """Sets the Sigma rule level."""
        self.sigma_level = level

    def set_description(self, description):
        """Set the Description property."""
        self.description = description

    def set_annotation(self, annotation):
        """Set the Annotation property."""
        self.annotation = annotation

    def set_generic_properties(self, fields):
        """Set the generic properties. """
        self.generic_properties = fields

    def _prefixed_tag(self):
        prefixes = {
            "Process.Start": "proc-start"
        }

        if self.event_type not in prefixes:
            return self.tag

        return "{}-{}".format(prefixes[self.event_type], self.tag)

    def __str__(self):
        """Builds and returns the [ActivityMonitoringRule] configuration block."""
        result = "[ActivityMonitoringRule]\n"

        # The Description is optional.
        if len(self.description) > 0:
            for description_line in self.description.splitlines():
                result += "# {}\n".format(description_line)

        # Make sure all required properties have at least a value that is somehow usable.
        if self.event_type is None:
            raise MalformedRuleException()

        if len(self.tag) == 0:
            raise MalformedRuleException()

        if len(self.name) == 0:
            raise MalformedRuleException()

        if len(self.query) == 0:
            raise MalformedRuleException()

        result += "RuleId = {}\n".format(self.id)
        result += "RuleName = {}\n".format(self.name)
        result += "EventType = {}\n".format(self.event_type)
        result += "Tag = {}\n".format(self._prefixed_tag())

        # The RiskScore is optional.
        # Set it, if a risk_score value is present.
        if self.risk_score > 0:
            result += "RiskScore = {}\n".format(self.risk_score)

        if len(self.annotation) > 0:
            result += "Annotation = {}\n".format(self.annotation)

        result += "Query = {}\n".format(self.query)

        if self.event_type == "Reg.Any":
            result += "Hive = HKLM,HKU\n"

        counter = 1
        for prop in self.generic_properties:

            # The following properties are included in all tagging events anyways.
            # There is no need to send them twice to the backend so we are ignoring them here.
            if prop in ["Process.Path", "Process.CommandLine", "Process.Name"]:
                continue
            # Generic properties are limited to 10.
            if counter > 10:
                break

            result += "GenericProperty{} = {}\n".format(counter, prop)
            counter += 1

        return result


def get_mitre_annotation_from_tag(tag):
    tag = tag.lower()
    if tag.startswith('attack.t'):
        return tag[7:].upper()
    return None


def get_annotation(tags):
    mitre_annotation_objects = []
    for tag in tags:
        mitre_annotation = get_mitre_annotation_from_tag(tag)
        if mitre_annotation is not None:
            mitre_annotation_objects.append(mitre_annotation)

    if len(mitre_annotation_objects) > 0:
        return json.dumps({ 'mitre_attack': mitre_annotation_objects })

    return ""


def get_parser_properties(sigmaparser):
    title = sigmaparser.parsedyaml['title']
    level = sigmaparser.parsedyaml['level']
    description = sigmaparser.parsedyaml['description']
    condition = sigmaparser.parsedyaml['detection']['condition']
    logsource = sigmaparser.parsedyaml['logsource']
    id = sigmaparser.parsedyaml['id']

    category = ''
    if 'category' in logsource:
        category = logsource['category'].lower()

    product = ''
    if 'product' in logsource:
        product = logsource['product'].lower()

    service = ''
    if 'service' in logsource:
        service = logsource['service'].lower()

    annotation = ''
    if 'tags' in sigmaparser.parsedyaml:
        annotation = get_annotation(sigmaparser.parsedyaml['tags'])

    return product, category, service, title, level, condition, description, annotation, id


def write_file_header(f, level):
    f.write("#\n")
    f.write("# The rules are generated from the Sigma GitHub repository at https://github.com/SigmaHQ/sigma\n")
    f.write("# Follow these steps to get the latest rules from the repository with Python\n")
    f.write("#    1. Clone the repository locally\n")
    f.write("#    2. Using a commandline, change working directory to the just cloned repository\n")
    f.write("#    3. Run sigmac -I --target uberagent -r rules/\n")
    f.write("#\n")
    f.write("# The rules in this file are marked with sigma-level: {}\n".format(level))
    f.write("#\n\n")


class uberAgentBackend(SingleTextQueryBackend):
    """Converts Sigma rule into uberAgent ESA's process tagging rules."""
    identifier = "uberagent"
    active = True
    config_required = False
    rule = None
    current_category = None
    recent_fields = []

    #
    # SingleTextQueryBackend
    #
    andToken = " and "
    orToken = " or "
    notToken = "not "
    subExpression = "(%s)"
    listExpression = "[%s]"
    listSeparator = ", "
    valueExpression = "\"%s\""
    nullExpression = "%s == ''"
    notNullExpression = "%s != ''"
    mapExpression = "%s == %s"
    mapListsSpecialHandling = True
    mapListValueExpression = "%s in %s"

    # Syntax for swapping wildcard conditions: Adding \ as escape character
    # Wildcard conditions are based on modifiers such as contains,
    # startswith, endswith
    mapWildcard = "%s like r%s"

    #
    # uberAgent field mapping
    #
    fieldMapping = {
        "commandline": "Process.CommandLine",
        "image": "Process.Path",
        "originalfilename": "Process.Name",
        "imageloaded": "Image.Path",
        "imagepath": "Image.Path",
        "parentcommandline": "Parent.CommandLine",
        "parentprocessname": "Parent.Name",
        "parentimage": "Parent.Path",
        "path": "Process.Path",
        "processcommandline": "Process.CommandLine",
        "command": "Process.CommandLine",
        "processname": "Process.Name",
        "user": "Process.User",
        "username": "Process.User",
        "company": "Process.Company"
    }

    fieldMappingPerCategory = {
        "process_creation": {
            "sha1": "Process.Hash.SHA1",
            "imphash": "Process.Hash.IMP",
            "childimage": "Process.Path"
            # Not yet supported.
            # "signed": "Process.IsSigned"
        },
        "image_load": {
            "sha1": "Image.Hash.SHA1",
            "imphash": "Image.Hash.IMP",
            "childimage": "Image.Path"
            # Not yet supported.
            # "signed": "Image.IsSigned"
        },
        "dns": {
            "query": "Dns.QueryRequest",
            "answer": "Dns.QueryResponse"
        },
        "dns_query": {
            "queryname": "Dns.QueryRequest",
        },
        "network_connection": {
            "destinationport": "Net.Target.Port",
            "destinationip": "Net.Target.Ip",
            "destinationhostname": "Net.Target.Name",
            "destinationisipv6": "Net.Target.IpIsV6",
            "sourceport": "Net.Source.Port"
        },
        "firewall": {
            "destination.port": "Net.Target.Port",
            "dst_ip": "Net.Target.Ip",
            "src_ip": "Net.Source.Ip"
        },
        "create_remote_thread": {
            "targetimage": "Process.Path",
            "startmodule": "Thread.StartModule",
            "startfunction": "Thread.StartFunctionName"
        },
        "registry_event": {
            "targetobject": "Reg.Key.Target",
            "newname": "Reg.Key.Path.New"
        },
        "registry_add": {
            "targetobject": "Reg.Key.Target",
            "newname": "Reg.Key.Path.New"
        },
        "registry_delete": {
            "targetobject": "Reg.Key.Target",
            "newname": "Reg.Key.Path.New"
        },
        "registry_set": {
            "targetobject": "Reg.Key.Target",
            "newname": "Reg.Key.Path.New"
        },
        "registry_rename": {
            "targetobject": "Reg.Key.Target",
            "newname": "Reg.Key.Path.New"
        } 
    }

    # We ignore some fields that we don't support yet but we don't want them to
    # throw errors in the console since we are aware of this.
    ignoreFieldList = [
        "description",
        "product",
        "logonid",
        "integritylevel",
        "currentdirectory",
        "parentintegritylevel",
        "eventid",
        "parentuser",
        "parent_domain",
        "signed",
        "parentofparentimage",
        "record_type",  # Related to network (DNS).
        "querystatus",  # Related to network (DNS).
        "initiated",  # Related to network connections. Seen as string 'true' / 'false'.
        "action",  # Related to firewall category.
        "targetprocessaddress",
        "sourceimage",
        "eventtype",
        "details"
    ]

    options = SingleTextQueryBackend.options + (
        ("exclusion", "", "List of separated GUIDs to execlude rule generation for.", None),
    )

    rules = []

    def trackRecentMappedField(self, field):
        if field not in self.recent_fields:
            self.recent_fields.append(field)


    def fieldNameMapping(self, fieldname, value):
        key = fieldname.lower()

        if self.current_category is not None:
            if self.current_category in self.fieldMappingPerCategory:
                if key in self.fieldMappingPerCategory[self.current_category]:
                    result = self.fieldMappingPerCategory[self.current_category][key]
                    self.trackRecentMappedField(result)
                    return result

        if key not in self.fieldMapping:
            if key in self.ignoreFieldList:
                raise IgnoreFieldException()
            else:
                raise NotImplementedError(
                    'The field name %s in category %s is not implemented.' % (fieldname, self.current_category))

        result = self.fieldMapping[key]
        self.trackRecentMappedField(result)
        return result

    def generateQuery(self, parsed):
        if parsed.parsedAgg:
            raise IgnoreAggregationException()

        return self.generateNode(parsed.parsedSearch)

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        product, category, service, title, level, condition, description, annotation, id = get_parser_properties(sigmaparser)

        # Do not generate a rule if the given category is unsupported by now.
        if not is_sigma_category_supported(category):
            return ""

        # Exclude all entries contained in backend configuration exclusion list.
        if id in self.backend_options["exclusion"]:
            return ""

        # We support windows rules and generic rules that don't have a specific product specifier - such as DNS.
        if product not in ["windows", ""]:
            return ""

        self.current_category = category

        try:
            rule = ActivityMonitoringRule()
            self.recent_fields = []

            query = super().generate(sigmaparser)
            if len(query) > 0:
                rule.set_id(id)
                rule.set_name(title)
                rule.set_tag(convert_sigma_name_to_uberagent_tag(title))
                rule.set_event_type(convert_sigma_category_to_uberagent_event_type(category))
                rule.set_query(query)
                rule.set_risk_score(convert_sigma_level_to_uberagent_risk_score(level))
                rule.set_sigma_level(level)
                rule.set_description(description)
                rule.set_annotation(annotation)
                rule.set_generic_properties(self.recent_fields)
                self.rules.append(rule)
                print("Generated rule <{}>.. [level: {}]".format(rule.name, level))
        except IgnoreTypedModifierException:
            return ""
        except IgnoreAggregationException:
            return ""
        except IgnoreFieldException:
            return ""
        except MalformedRuleException:
            return ""

    def serialize_file(self, name, level):
        count = 0
        with open(name, "w", encoding='utf8') as file:
            write_file_header(file, level)
            for rule in self.rules:
                try:
                    serialized_rule = str(rule)
                    if rule.sigma_level == level:
                        file.write(serialized_rule + "\n")
                        count = count + 1
                except MalformedRuleException:
                    continue
            file.close()
        return count

    def finalize(self):
        count_critical = self.serialize_file("uberAgent-ESA-am-sigma-critical.conf", "critical")
        count_high = self.serialize_file("uberAgent-ESA-am-sigma-high.conf", "high")
        count_low = self.serialize_file("uberAgent-ESA-am-sigma-low.conf", "low")
        count_medium = self.serialize_file("uberAgent-ESA-am-sigma-medium.conf", "medium")
        print("Generated {} activity monitoring rules..".format(len(self.rules)))
        print(
            "This includes {} critical rules, {} high rules, {} medium rules and {} low rules..".format(count_critical,
                                                                                                        count_high,
                                                                                                        count_medium,
                                                                                                        count_low))

        print("There are %d unsupported categories." % len(gUnsupportedCategories))
        for category in gUnsupportedCategories:
            print("Category %s has %d unsupported rules." % (category, gUnsupportedCategories[category]))

    def generateTypedValueNode(self, node):
        raise IgnoreTypedModifierException()

    def generateMapItemTypedNode(self, fieldname, value):
        raise IgnoreTypedModifierException()

    def generateMapItemListNode(self, key, value):
        return "(" + (" or ".join([self.mapWildcard % (key, self.generateValueNode(item)) for item in value])) + ")"

    def generateMapItemNode(self, node):
        fieldname, value = node
        transformed_fieldname = self.fieldNameMapping(fieldname, value)

        if value is None:
            return self.nullExpression % (transformed_fieldname,)

        has_wildcard = re.search(r"((\\(\*|\?|\\))|\*|\?|_|%)", self.generateNode(value))

        if "," in self.generateNode(value) and not has_wildcard:
            return self.mapListValueExpression % (transformed_fieldname, self.generateNode(value))
        elif type(value) == list:
            return self.generateMapItemListNode(transformed_fieldname, value)
        elif self.mapListsSpecialHandling == False and type(value) in (
                str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
            if has_wildcard:
                return self.mapWildcard % (transformed_fieldname, self.generateNode(value))
            else:
                return self.mapExpression % (transformed_fieldname, self.generateNode(value))
        elif has_wildcard:
            return self.mapWildcard % (transformed_fieldname, self.generateNode(value))
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def cleanValue(self, val):
        if not isinstance(val, str):
            return str(val)

        # Single backlashes which are not in front of * or ? are doubled
        val = re.sub(r"(?<!\\)\\(?!(\\|\*|\?))", r"\\\\", val)

        # Replace _ with \_ because _ is a sql wildcard
        val = re.sub(r'_', r'\_', val)

        # Replace % with \% because % is a sql wildcard
        val = re.sub(r'%', r'\%', val)

        # Replace " with \" because " is a string literal symbol and must be escaped
        val = re.sub(r'"', r'\"', val)

        # Replace * with %, if even number of backslashes (or zero) in front of *
        val = re.sub(r"(?<!\\)(\\\\)*(?!\\)\*", r"\1%", val)

        # Replace ? with _, if even number of backslashes (or zero) in front of ?
        val = re.sub(r"(?<!\\)(\\\\)*(?!\\)\?", r"\1_", val)
        return val



# Output backends for sigmac
# Copyright 2021 Lacework, Inc.
# Authors:
# David Hazekamp (david.hazekamp@lacework.net)
# Rachel Rice (rachel.rice@lacework.net)

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import re
import textwrap
import yaml

import sigma

from sigma.backends.base import SingleTextQueryBackend
from sigma.backends.exceptions import BackendError
from sigma.parser.condition import NodeSubexpression, ConditionOR
from sigma.parser.modifiers.base import SigmaTypeModifier


LACEWORK_CONFIG = yaml.load(
    textwrap.dedent('''
    ---
    version: 0.4
    services:
      cloudtrail:
        source: CloudTrailRawEvents
        fieldMap:
          - sigmaField: eventName
            laceworkField: EVENT_NAME
            matchType: exact
            continue: false
          - sigmaField: eventSource
            laceworkField: EVENT_SOURCE
            matchType: exact
            continue: false
          - sigmaField: errorCode
            laceworkField: ERROR_CODE
            matchType: exact
            continue: false
          - sigmaField: "^(.*)$"
            laceworkField: EVENT:$1
            matchType: regex
            continue: true
          - sigmaField: "^(.*?)\\\\.type$"
            laceworkField: '$1."type"'
            matchType: regex
            continue: true
        returns:
          - INSERT_ID
          - INSERT_TIME
          - EVENT_TIME
          - EVENT
        alertProfile: LW_CloudTrail_Alerts
    product.categories:
      linux.file_create:
        sources:
          - LW_HE_FILES
        conditions:
          # evaluated hourly and file create time within the last hour
          - and diff_minutes(FILE_CREATED_TIME, current_timestamp_sec()::timestamp) <= 60
        fieldMap:
          - sigmaField: TargetFilename
            laceworkField: PATH
            matchType: exact
        returns:
          - RECORD_CREATED_TIME
          - MID
          - PATH
          - FILE_TYPE
          - SIZE
          - FILEDATA_HASH
          - OWNER_UID
          - OWNER_USERNAME
          - FILE_CREATED_TIME
        alertProfile: LW_HE_FILES_DEFAULT_PROFILE.HE_File_NewViolation
      linux.network_connection:
        sources:
          - LW_HA_CONNECTION_SUMMARY as HACM
          - array_to_rows(ENDPOINT_DETAILS) as (EP_PA)
        fieldMap:
          - sigmaField: Image
            laceworkField: EXE_PATH
            matchType: exact
            action: selfjoin
            selfJoinFilter: HACM.SRC_ENTITY_TYPE = 'Process' AND (HACM.SRC_ENTITY_ID:mid::NUMBER, HACM.SRC_ENTITY_ID:pid_hash::NUMBER) IN { source { LW_HE_PROCESSES AS HEP } filter { EXE_PATH like '%/bin/bash' } return { HEP.MID, HEP.PID_HASH }}
          - sigmaField: DestinationIp
            laceworkField: EP_PA:dst_ip_addr
            matchType: exact
          - sigmaField: DestinationHostname
            laceworkField:
            matchType: exact
            action: raise
        returns:
          - HACM.BATCH_END_TIME
          - HACM.BATCH_START_TIME
          - HACM.DST_ENTITY_ID
          - HACM.DST_ENTITY_TYPE
          - HACM.DST_IN_BYTES
          - HACM.DST_OUT_BYTES
          - HACM.ENDPOINT_DETAILS
          - HACM.NUM_CONNS
          - HACM.SRC_ENTITY_ID
          - HACM.SRC_ENTITY_TYPE
          - HACM.SRC_IN_BYTES
          - HACM.SRC_OUT_BYTES
        alertProfile: LW_HA_CONNECTION_SUMMARY_DEFAULT_PROFILE.HA_Connection_Violation
      linux.process_creation:
        sources:
          - LW_HE_PROCESSES
        conditions:
          # evaluated hourly and file create time within the last hour
          - and diff_minutes(PROCESS_START_TIME, current_timestamp_sec()::timestamp) <= 60
        fieldMap:
          - sigmaField: ParentImage
            laceworkField: EXE_PATH
            matchType: exact
            action: selfjoin
            selfJoinFilter: (MID, PPID_HASH) IN { source { LW_HE_PROCESSES } filter { $query$ } return { MID, PID_HASH } }
          - sigmaField: Image
            laceworkField: EXE_PATH
            matchType: exact
          - sigmaField: ParentCommandLine
            laceworkField: CMDLINE
            matchType: exact
            action: selfjoin
            selfJoinFilter: (MID, PPID_HASH) IN { source { LW_HE_PROCESSES } filter { $query$ } return { MID, PID_HASH } }
          - sigmaField: CommandLine
            laceworkField: CMDLINE
            matchType: exact
          - sigmaField: CurrentDirectory
            laceworkField: CWD
            matchType: exact
          - sigmaField: LogonId
            laceworkField:
            matchType: exact
            action: ignore
          - sigmaField: User
            laceworkField: USERNAME
            matchType: exact
        returns:
          - RECORD_CREATED_TIME
          - MID
          - PID
          - EXE_PATH
          - CMDLINE
          - CWD
          - ROOT
          - USERNAME
          - PROCESS_START_TIME
        alertProfile: LW_HE_PROCESSES_DEFAULT_PROFILE.HE_Process_NewViolation
    '''),
    Loader=yaml.SafeLoader
)


def safe_get(obj, name, inst):
    """
    Sweet helper for getting objects
    """
    try:
        assert isinstance(obj[name], inst)
        value = obj[name]
    except Exception:
        value = inst()

    return value


def get_output_format(config):
    return (
        'json'
        if (
            safe_get(config, 'json', bool)
            or safe_get(config, 'JSON', bool)
        )
        else 'yaml'
    )


# YAML Tools
def str_presenter(dumper, data):
    if len(data.splitlines()) > 1:  # check for multiline string
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)


def none_representer(dumper, _):
    return dumper.represent_scalar(u'tag:yaml.org,2002:null', '')


yaml.add_representer(str, str_presenter)
yaml.add_representer(type(None), none_representer)


class LaceworkBackend(SingleTextQueryBackend):
    """Converts Sigma rule into Lacework Policy Platform"""
    identifier = "lacework"
    active = True
    # our approach to config will be such that we support both an
    # embedded or specified config.
    config_required = False

    andToken = ' and '
    orToken = ' or '
    notToken = 'not '
    subExpression = '(%s)'
    listExpression = 'in (%s)'
    listSeparator = ', '
    valueExpression = "'%s'"
    nullExpression = '%s is null'
    notNullExpression = '%s is not null'
    mapExpression = '%s = %s'
    mapListValueExpression = '%s %s'
    reEscape = re.compile("(')")

    def generate(self, sigmaparser):
        """
        Method is called for each sigma rule and receives the parsed rule (SigmaParser)
        """
        # 1. get embedded config global
        config = LACEWORK_CONFIG

        # 2. overlay backend options
        config.update(self.backend_options)

        # 3. set a class instance variable for sigma fields
        self.laceworkSigmaFields = LaceworkQuery.get_fields(sigmaparser)

        # 4. set a class instance variable for lacework field mapping
        self.laceworkFieldMap = LaceworkQuery.get_field_map(LACEWORK_CONFIG, sigmaparser)

        # 5. get output format
        output_format = get_output_format(config)

        # determine if we're generating query/policy/both
        result = ''
        if LaceworkQuery.should_generate_query(config):
            query = LaceworkQuery(
                config, sigmaparser, self, output_format=output_format)
            result += str(query)
        if LaceworkPolicy.should_generate_policy(config):
            policy = LaceworkPolicy(
                config, sigmaparser, output_format=output_format)

            # if we're in json mode and have already generated a query
            # add a newline before emitting policy
            if result and output_format == 'json':
                result += '\n'

            result += str(policy)

        return result

    def generateNode(self, node):
        if type(node) == sigma.parser.condition.ConditionAND:
            return self.applyOverrides(self.generateANDNode(node))
        elif type(node) == sigma.parser.condition.ConditionOR:
            return self.applyOverrides(self.generateORNode(node))
        elif type(node) == sigma.parser.condition.ConditionNOT:
            return self.applyOverrides(self.generateNOTNode(node))
        elif type(node) == sigma.parser.condition.ConditionNULLValue:
            return self.applyOverrides(self.generateNULLValueNode(node))
        elif type(node) == sigma.parser.condition.ConditionNotNULLValue:
            return self.applyOverrides(self.generateNotNULLValueNode(node))
        elif type(node) == sigma.parser.condition.NodeSubexpression:
            return self.applyOverrides(self.generateSubexpressionNode(node))
        elif type(node) == tuple:
            return self.applySelfJoinFilter(node, self.applyOverrides(self.generateMapItemNode(node)))
        elif type(node) in (str, int):
            return self.applyOverrides(self.generateValueNode(node))
        elif type(node) == list:
            return self.applyOverrides(self.generateListNode(node))
        elif isinstance(node, SigmaTypeModifier):
            return self.applyOverrides(self.generateTypedValueNode(node))
        else:
            raise TypeError("Node type %s was not expected in Sigma parse tree" % (str(type(node))))

    def applySelfJoinFilter(self, node, query):
        if type(node) != tuple:
            raise NotImplementedError('selfJoinFilter is not wired up for node type %s' % (str(type(node))))
        fieldname, value = node

        sjf = self._get_self_join_filter(fieldname)
        if not sjf:
            return query
        return sjf.replace('$query$', query)

    def generateValueNode(self, node):
        """
        Value Expression for Lacework Query Language (LQL)

        If value is a field name
        1.  Do not wrap in valueExpression
        2.  Transform using fieldNameMapping()
        """
        node = self.cleanValue(str(node).strip())

        if node in self.laceworkSigmaFields:
            if self._should_ignore_field(node):
                return None
            return self.fieldNameMapping(node, None)
        return self.valueExpression % node

    def generateMapItemNode(self, node):
        """
        Map Expression for Lacework Query Language (LQL)

        Special handling for contains by inspecting value for wildcards
        """
        fieldname, value = node

        if self._should_ignore_field(fieldname):
            return None
        transformed_fieldname = self.fieldNameMapping(fieldname, value)

        # is not null
        if value == '*':
            if ':' in transformed_fieldname:
                return f'value_exists({transformed_fieldname})'
            return f'{transformed_fieldname} is not null'
        # contains
        if (
            isinstance(value, str)
            and value.startswith('*')
            and value.endswith('*')
        ):
            value = self.generateValueNode(value[1:-1])
            return f"contains({transformed_fieldname}, {value})"
        # startswith
        if (
            isinstance(value, str)
            and value.endswith('*')  # a wildcard at the end signifies startswith
        ):
            value = self.generateValueNode(value[:-1])
            return f"starts_with({transformed_fieldname}, {value})"
        # endswith
        if (
            isinstance(value, str)
            and value.startswith('*')  # a wildcard at the start signifies endswith
        ):
            value = f'%{value[1:]}'
            new_value = self.generateValueNode(value)
            if new_value != (self.valueExpression % value):
                raise BackendError(
                    'Lacework backend only supports endswith for literal string values')
            return f"{transformed_fieldname} LIKE {new_value}"
        if isinstance(value, (str, int)):
            return self.mapExpression % (transformed_fieldname, self.generateNode(value))
        # mapListsHandling
        elif type(value) == list:
            # if a list contains values with wildcards we can't use standard handling ("in")
            if any([x for x in value if x.startswith('*') or x.endswith('*')]):
                node = NodeSubexpression(
                    ConditionOR(None, None, *[(transformed_fieldname, x) for x in value])
                )
                return self.generateNode(node)
            return self.generateMapItemListNode(transformed_fieldname, value)
        elif value is None:
            return self.nullExpression % (transformed_fieldname, )
        else:
            raise TypeError(
                f'Lacework backend does not support map values of type {type(value)}')

    def _should_ignore_field(self, fieldname):
        """
        Whether to ignore field for Lacework Query Language (LQL)
        """
        if not (isinstance(fieldname, str) and fieldname):
            return False

        for map in self.laceworkFieldMap:
            if not isinstance(map, dict):
                continue

            sigma_field = safe_get(map, 'sigmaField', str)
            if not sigma_field:
                continue

            # ignore
            if (
                map.get('matchType') == 'exact'
                and sigma_field == fieldname
                and map.get('action') == 'ignore'
            ):
                return True

        return False

    def _get_self_join_filter(self, fieldname):
        """
        Whether we're implementing a self-join filter within Lacework Query Language (LQL)
        """
        if not (isinstance(fieldname, str) and fieldname):
            return None

        for map in self.laceworkFieldMap:
            if not isinstance(map, dict):
                continue

            sigma_field = safe_get(map, 'sigmaField', str)
            if not sigma_field:
                continue

            sjf = map.get('selfJoinFilter')

            # self join filter
            if (
                map.get('action') == 'selfjoin'
                and sigma_field == fieldname
                and sjf and isinstance(sjf, str)
            ):
                return sjf

        return None

    @staticmethod
    def _check_unsupported_field(action, fieldname):
        if action == 'raise':
            raise BackendError(
                f'Lacework backend does not support the {fieldname} field')

    def fieldNameMapping(self, fieldname, value):
        """
        Field Name Mapping for Lacework Query Language (LQL)

        The Lacework backend is not using a traditional config.
        As such we map field names here using our custom backend config.
        """
        if not (isinstance(fieldname, str) and fieldname):
            return fieldname

        for map in self.laceworkFieldMap:
            if not isinstance(map, dict):
                continue

            sigma_field = safe_get(map, 'sigmaField', str)
            if not sigma_field:
                continue

            lacework_field = safe_get(map, 'laceworkField', str)
            if (not lacework_field) and map.get('action') != 'raise':
                continue

            continyu = safe_get(map, 'continue', bool)

            # exact
            if (
                map.get('matchType') == 'exact'
                and sigma_field == fieldname
            ):
                self._check_unsupported_field(map.get('action'), fieldname)
                fieldname = lacework_field
                if not continyu:
                    return fieldname

            # startswith
            if (
                map.get('matchType') == 'startswith'
                and fieldname.startswith(sigma_field)
            ):
                self._check_unsupported_field(map.get('action'), fieldname)
                fieldname = f'{lacework_field}{fieldname[len(sigma_field):]}'
                if not continyu:
                    return fieldname

            # regex
            if map.get('matchType') == 'regex':
                fieldname_re = re.compile(sigma_field)
                fieldname_match = fieldname_re.match(fieldname)

                if not fieldname_match:
                    continue

                self._check_unsupported_field(map.get('action'), fieldname)

                for i, group in enumerate(fieldname_match.groups(), start=1):
                    if group is None:
                        continue
                    fieldname = lacework_field.replace(f'${i}', group)

                if not continyu:
                    return fieldname

        return fieldname


class LaceworkQuery:

    def __init__(
        self,
        config,
        sigmaparser,
        backend,
        output_format='yaml'
    ):
        rule = sigmaparser.parsedyaml
        conditions = sigmaparser.condparsed

        # 0. Get Output Format
        self.output_format = str(output_format).lower()

        # 1. Get Logsource
        self.logsource_type, self.logsource_name = self.get_logsource(rule)

        # 2. Get Logsource Config
        self.logsource_config = self.get_logsource_config(
            config, self.logsource_type, self.logsource_name)

        # 3. Get Query ID
        self.title, self.query_id = self.get_query_id(rule)

        # 4. Get Query Source
        self.query_sources = self.get_query_sources(
            self.logsource_name, self.logsource_config)

        # 5. Get Query Returns
        self.returns = self.get_query_returns(
            self.logsource_name, self.logsource_config)

        # 6. Get Query Text
        self.query_text = self.get_query_text(backend, conditions)

    def get_query_text(self, backend, rule_conditions):
        query_template = (
            '{id} {{\n'
            '    {source_block}\n'
            '    {filter}\n'
            '    {return_block}\n'
            '}}'
        )

        # 1. get_query_source_block
        source_block = self.get_query_source_block()

        # 2. get_query_filters
        config_conditions = safe_get(self.logsource_config, 'conditions', list)
        filter_block = self.get_query_filter_block(backend, rule_conditions, config_conditions)

        # 3. get_query_returns
        return_block = self.get_query_return_block()

        return query_template.format(
            id=self.query_id,
            source_block=source_block,
            filter=filter_block,
            return_block=return_block
        )

    def get_query_source_block(self):
        source_block_template = (
            'source {{\n'
            '        {source}\n'
            '    }}'
        )

        return source_block_template.format(
            source=',\n        '.join(self.query_sources)
        )

    def get_query_return_block(self):
        return_block_template = (
            'return distinct {{\n'
            '{returns}\n'
            '    }}'
        )
        return return_block_template.format(
            returns=',\n'.join(f'        {r}' for r in self.returns)
        )

    def __iter__(self):
        for key, attr in {
            'queryId': 'query_id',
            'queryText': 'query_text'
        }.items():
            yield (key, getattr(self, attr))

    def __str__(self):
        o = dict(self)

        if self.output_format == 'json':
            return json.dumps(o, indent=4)

        return yaml.dump(
            o,
            explicit_start=True,
            default_flow_style=False,
            sort_keys=False
        )

    @staticmethod
    def get_fields(sigmaparser):
        return safe_get(sigmaparser.parsedyaml, 'fields', list)

    @staticmethod
    def get_field_map(config, sigmaparser):
        logsource_type, logsource_name = LaceworkQuery.get_logsource(sigmaparser.parsedyaml)
        logsource_config = LaceworkQuery.get_logsource_config(config, logsource_type, logsource_name)

        return safe_get(logsource_config, 'fieldMap', list)

    @staticmethod
    def should_generate_query(backend_options):
        # if we are explicitly requesting a query
        if (
            'query' in backend_options
            and backend_options['query'] is True
        ):
            return True
        # if we are explicitly requesting a policy
        if (
            'policy' in backend_options
            and backend_options['policy'] is True
        ):
            return False
        # we're not being explicit about anything
        return True

    @staticmethod
    def get_logsource(rule):
        logsource = safe_get(rule, 'logsource', dict)
        if 'service' in logsource:
            return 'services', logsource['service']
        if {'product', 'category'}.issubset(set(logsource)):
            return 'product.categories', f"{logsource['product']}.{logsource['category']}"
        return 'unknown', 'unknown'

    @staticmethod
    def get_logsource_config(config, logsource_type, logsource_name):
        config = safe_get(config, logsource_type, dict)
        logsource_config = safe_get(config, logsource_name, dict)

        # 1. validate logsource service
        if not logsource_config:
            raise BackendError(
                f'Log source {logsource_name} is not supported by the Lacework backend')

        return logsource_config

    @staticmethod
    def get_query_id(rule):
        title = safe_get(rule, 'title', str) or 'Unknown'
        # TODO: might need to replace additional non-word characters
        query_id = f'Sigma_{title}'.replace(" ", "_").replace("/", "_Or_").replace("-", "_")

        return title, query_id

    @staticmethod
    def get_query_sources(logsource_name, logsource_config):
        # 4. validate service has a source mapping
        source = safe_get(logsource_config, 'source', str)
        sources = safe_get(logsource_config, 'sources', list)

        if sources:
            return sources
        elif source:
            return [source]

        raise BackendError(
            f'Lacework backend could not determine source for logsource {logsource_name}')

    @staticmethod
    def get_query_returns(logsource_name, logsource_config):
        returns = safe_get(logsource_config, 'returns', list)

        if not returns:
            raise BackendError(
                f'Lacework backend could not determine returns for logsource {logsource_name}')

        return returns

    @staticmethod
    def get_query_filter_block(backend, rule_conditions, config_conditions):
        filter_block_template = (
            'filter {{\n'
            '        {filter}\n'
            '    }}'
        )

        for parsed in rule_conditions:
            query = backend.generateQuery(parsed)
            before = backend.generateBefore(parsed)
            after = backend.generateAfter(parsed)

            filter = ""
            if before is not None:
                filter = before
            if query is not None:
                filter += query
            if after is not None:
                filter += after

            if config_conditions:
                filter += f" {' '.join(config_conditions)}"
            return filter_block_template.format(filter=filter)


class LaceworkPolicy:
    def __init__(
        self,
        config,
        sigmaparser,
        output_format='yaml'
    ):
        rule = sigmaparser.parsedyaml

        # 0. Get Output Format
        self.output_format = str(output_format).lower()

        # 1. Get Service Name
        self.logsource_type, self.logsource_name = LaceworkQuery.get_logsource(rule)

        # 2. Get Service Config
        self.logsource_config = LaceworkQuery.get_logsource_config(
            config, self.logsource_type, self.logsource_name)

        # 3. Get Title
        # 4. Get Query ID
        self.title, self.query_id = LaceworkQuery.get_query_id(rule)

        # 5. Get Enabled
        self.enabled = False

        # 6. Get Policy Type
        self.policy_type = 'Violation'

        # 7. Get Alert Enabled
        self.alert_enabled = False

        # 8. Get Alert Profile
        self.alert_profile = self.get_alert_profile(
            self.logsource_name, self.logsource_config)

        # 9. Get Limit
        self.limit = 1000

        # 10. Get Severity
        self.severity = safe_get(rule, 'level', str) or 'medium'

        # 11. Get Description
        self.description = safe_get(rule, 'description', str)

        # 12. Get Remediation
        self.remediation = 'Remediation steps are not represented in Sigma rule specification'

    def __iter__(self):
        for key, attr in {
            'title': 'title',
            'enabled': 'enabled',
            'policyType': 'policy_type',
            'alertEnabled': 'alert_enabled',
            'alertProfile': 'alert_profile',
            'queryId': 'query_id',
            'limit': 'limit',
            'severity': 'severity',
            'description': 'description',
            'remediation': 'remediation'
        }.items():
            yield (key, getattr(self, attr))

    def __str__(self):
        o = dict(self)

        if self.output_format == 'json':
            return json.dumps(o, indent=4)

        return yaml.dump(
            o,
            explicit_start=True,
            default_flow_style=False,
            sort_keys=False
        )

    @staticmethod
    def should_generate_policy(backend_options):
        # if we are explicitly requesting a query
        if (
            'policy' in backend_options
            and backend_options['policy'] is True
        ):
            return True
        # if we are explicitly requesting a policy
        if (
            'query' in backend_options
            and backend_options['query'] is True
        ):
            return False
        # we're not being explicit about anything
        return True

    @staticmethod
    def get_alert_profile(logsource_name, logsource_config):
        alert_profile = safe_get(logsource_config, 'alertProfile', str)

        if not alert_profile:
            raise BackendError(
                f'Lacework backend could not determine alert profile for logsource {logsource_name}')

        return alert_profile



import re
from .base import SingleTextQueryBackend
import json

class LogiqBackend(SingleTextQueryBackend):
    """Converts Sigma rule into LOGIQ event rule api payload """
    identifier = "logiq"
    config_required = False
    active = True
    reEscape = re.compile('(")')
    reClear = None
    andToken = " && "
    orToken = " || "
    notToken = " !~ "
    subExpression = "%s"
    listExpression = "%s"
    listSeparator = ", "
    valueExpression = "message =~ \'%s\'"
    keyExpression = "%s"
    nullExpression = "!~ %s"
    notNullExpression = "!%s"
    mapExpression = "(%s=%s)"
    mapListsSpecialHandling = True

    reEscape = re.compile("([\\|()\[\]{}.^$+])")

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""

        eventRule = dict()
        eventRule["name"] = sigmaparser.parsedyaml["title"]
        eventRule["groupName"] = sigmaparser.parsedyaml["logsource"].get("product", "")
        eventRule["description"] = sigmaparser.parsedyaml["description"]
        eventRule["condition"] = sigmaparser.parsedyaml["detection"]
        eventRule["level"] = sigmaparser.parsedyaml["level"]

        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed)
            before = self.generateBefore(parsed)
            after = self.generateAfter(parsed)

            eventRule["condition"] = ""
            if before is not None:
                eventRule["condition"] = before
            if query is not None:
                eventRule["condition"] += query
            if after is not None:
                eventRule["condition"] += after

            return json.dumps(eventRule)

    def cleanValue(self, val):
        if val.startswith('*'):
            val = val.replace("*","/*")
      
        return val

    def generateListNode(self, node):
        if not set([type(value) for value in node]).issubset({str, int}):
            raise TypeError("List values must be strings or numbers")
        return self.generateORNode(node)



# Output backends for sigmac
# Copyright 2016-2018 Thomas Patzke

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .base import BaseBackend

### Backends for development purposes

class FieldnameListBackend(BaseBackend):
    """List all fieldnames from given Sigma rules for creation of a field mapping configuration."""
    identifier = "fieldlist"
    active = True
    config_required = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields = set()

    def generateQuery(self, parsed):
        fields = list(flatten(self.generateNode(parsed.parsedSearch)))
        if parsed.parsedAgg:
            fields += self.generateAggregation(parsed.parsedAgg)
        self.fields.update(fields)

    def generateANDNode(self, node):
        return [self.generateNode(val) for val in node]

    def generateORNode(self, node):
        return self.generateANDNode(node)

    def generateNOTNode(self, node):
        return self.generateNode(node.item)

    def generateSubexpressionNode(self, node):
        return self.generateNode(node.items)

    def generateListNode(self, node):
        if not set([type(value) for value in node]).issubset({str, int}):
            raise TypeError("List values must be strings or numbers")
        return [self.generateNode(value) for value in node]

    def generateMapItemNode(self, node):
        key, value = node
        return [key]

    def generateValueNode(self, node):
        return []

    def generateNULLValueNode(self, node):
        return [node.item]

    def generateNotNULLValueNode(self, node):
        return [node.item]

    def generateAggregation(self, agg):
        fields = list()
        if agg.groupfield is not None:
            fields.append(agg.groupfield)
        if agg.aggfield is not None:
            fields.append(agg.aggfield)
        return fields

    def finalize(self):
        return "\n".join(sorted(self.fields))

# Helpers
def flatten(l):
  for i in l:
      if type(i) == list:
          yield from flatten(i)
      else:
          yield i



import re
from datetime import datetime

import sigma
from sigma.backends.base import SingleTextQueryBackend
from sigma.backends.mixins import MultiRuleOutputMixin

from .exceptions import NotSupportedError
from ..parser.condition import SigmaAggregationParser
from ..parser.modifiers.base import SigmaTypeModifier
from ..parser.modifiers.transform import SigmaContainsModifier, SigmaStartswithModifier, SigmaEndswithModifier
from ..parser.modifiers.type import SigmaRegularExpressionModifier

comparative = ["greater_than",
               "greater_equal",
               "less_than",
               "less_equal",
               ]

class ChronicleBackend(SingleTextQueryBackend):
    """Converts Sigma rule into Google Chronicle YARA-L. Contributed by SOC Prime. https://socprime.com"""
    identifier = "chronicle"
    active = True
    andToken = " and "
    #\\\
    reEscape = re.compile('([\"]|(\\\\))')
    reClear = re.compile('`')

    orToken = " or "
    notToken = "not "
    subExpression = "(%s)"
    valueExpression = "\"%s\""
    mapExpression = "%s = %s"
    listExpression = "(%s)"
    listSeparator = " or "
    config_required = True
    mapListsSpecialHandling = True

    def __init__(self, *args, **kwargs):
        self.defaultEventName = "event"
        self.condition_name = None
        self.parsed_detection = None
        self.author = None
        self.description = None
        self.created = None
        self.title = None
        self.references = None
        self.rule_count = 0
        return super().__init__(*args, **kwargs)

    def cleanValue(self, val):
        if val and isinstance(val, str) and val.endswith("/"):
            val = val.rstrip("/")
        if val and isinstance(val, str) and val.startswith("\\"):
            val = val.lstrip("\\")
        return super().cleanValue(val)

    def parseTitle(self, title):
        new_title = re.sub(re.compile('[()*:;+!,\[\].?"-/]'), "", title.lower())
        new_title = re.sub(re.compile('\s'), "_", new_title.lower())
        index = 0
        for i, title_char in enumerate(new_title):
            if not title_char.isdigit():
                index = i
                break
        new_title = new_title[index:]
        new_title = new_title.strip("_")
        return new_title

    def generateMapItemNode(self, node):
        fieldname, value = node

        transformed_fieldname = self.fieldNameMapping(fieldname, value)
        if type(value) in (str, int):
            return self.regex_check(transformed_fieldname=transformed_fieldname, val=value)
        elif type(value) == list:
            return self.generateMapItemListNode(transformed_fieldname, value)
        elif isinstance(value, SigmaTypeModifier):
            return self.generateMapItemTypedNode(transformed_fieldname, value)
        elif value is None:
            return self.nullExpression % (transformed_fieldname, )
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def createFinalRule(self, body):
        # Spaces required in rule for structure
        function_name = self.parseTitle(self.title)
        if self.rule_count != 0:
            function_name += "_part_{}".format(self.rule_count)

        meta = """ meta:\n    author = \"{author}\"\n    description = \"{description}\"\n    reference = \"{reference}\"\n    version = \"0.01\"""".format(
            author=self.author, description=self.description, reference=""
        )
        if self.created:
            meta += "\n    created = \"{}\"".format(self.created)
        if any(self.logsource):
            logsources = "\n    ".join([f'{i} = "{j}"' for i, j in self.logsource.items() if i not in ("description", "definition")])
            meta += "\n    {}".format(logsources)
        if self.tags:
            tags = ", ".join([item.replace("attack.", "") for item in self.tags])
            meta += "\n    mitre = \"{}\"".format(tags)
        condition_func = """  condition:\n    {condition}""".format(condition=self.condition)
        result = """rule {function_name} {{\n{meta}\n\n  events:\n{function}\n\n{condition}\n}}""".format(
            function_name=function_name,
            meta=meta,
            function=body,
            condition=condition_func
        )
        self.rule_count += 1
        return result

    def fieldNameMapping(self, fieldname, value):
        return f"${self.condition_name}.{fieldname}"

    def regex_check(self, transformed_fieldname, val):
        if val and isinstance(val, str) and '*' in val:
            val = val.replace("\*", "*")
            val = self.cleanValue(val)
            val = val.replace("(", "\(")
            val = val.replace(")", "\)")
            val = re.compile(r'([+.?])').sub("\\\\\g<1>", val)
            val = val.replace("*", ".*")
            return f"re.regex({transformed_fieldname}, `{val}`)"
        if val and isinstance(val, str):
            return self.mapExpression % (transformed_fieldname, self.generateNode(val))
        else:
            return self.mapExpression % (transformed_fieldname, self.generateNode(val))

    def generateMapItemListNode(self, fieldname, value):
        list_query = []
        for item in value:
            updated_field_value = self.regex_check(transformed_fieldname=fieldname, val=item)
            list_query.append(updated_field_value)
        if len(list_query) > 1:
            return "(" + " or ".join(list_query) + ")"
        return list_query[0]

    def generate(self, sigmaparser):
        detection = sigmaparser.parsedyaml.get("detection")
        condition_name = [item for item in detection.keys() if item not in ("condition", "keywords")]
        if any(condition_name):
            self.condition_name = condition_name[0]
        else:
            self.condition_name = "event"
        self.author = sigmaparser.parsedyaml.get("author")
        self.title = sigmaparser.parsedyaml.get("title")
        description = "{} Author: {}.".format(sigmaparser.parsedyaml.get("description"), self.author)
        description = description.replace("\\", "\\\\")
        description = description.replace("\n", "")
        self.description = description.replace('"', '\\"')
        self.created = sigmaparser.parsedyaml.get("date", datetime.now().strftime("%Y-%m-%d"))
        references = sigmaparser.parsedyaml.get("reference", [])
        if not any(references):
            references = sigmaparser.parsedyaml.get("references", [])
        self.references = references
        self.logsource = sigmaparser.parsedyaml.get("logsource") if sigmaparser.parsedyaml.get("logsource") else sigmaparser.parsedyaml.get("logsources", {})
        self.tags = sigmaparser.parsedyaml.get("tags")
        for parsed in sigmaparser.condparsed:
            aggregation = None
            translate = self.generateQuery(parsed)
            self.condition = "${}".format(self.condition_name)
            if parsed.parsedAgg:
                translate = self.generateAggregation(agg=parsed.parsedAgg, body=translate)
            return self.createFinalRule(body=translate)

    def generateQuery(self, parsed):
        result = self.generateNode(parsed.parsedSearch)
        return result

    def generateAggregation(self, agg, body):
        if agg is None:
            return ""
        if agg.aggfunc == SigmaAggregationParser.AGGFUNC_NEAR:
            raise NotImplementedError(
                "The 'near' aggregation operator is not "
                + f"implemented for the %s backend" % self.identifier
            )
        if agg.aggfunc_notrans != 'count' and agg.aggfield is None:
            raise NotSupportedError(
                "The '%s' aggregation operator " % agg.aggfunc_notrans
                + "must have an aggregation field for the %s backend" % self.identifier
            )
        if agg.aggfunc_notrans == 'count':
            if agg.groupfield:
                self.condition = "${condition} and #target {op} {cond}".format(condition=self.condition_name,
                                                                                          field=agg.groupfield,
                                                                                          op=agg.cond_op,
                                                                                          cond=agg.condition)
                body += "\n${condition}.{field} = $target".format(condition=self.condition_name, field=agg.groupfield,)
            else:
                self.condition = "#{} {} {}".format(self.condition_name, agg.cond_op, agg.condition)
            return body


# Output backends for sigmac
# Copyright 2020 FireEye, Inc.
# Author: Alek Rollyson <alek.rollyson@fireeye.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sigma.backends.base import SingleTextQueryBackend

from sigma.parser.modifiers.base import SigmaTypeModifier
from sigma.parser.modifiers.transform import (
    SigmaContainsModifier,
    SigmaStartswithModifier,
    SigmaEndswithModifier,
)
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier


class FireEyeHelixBackend(SingleTextQueryBackend):
    """Converts Sigma rule into FireEye Helix Query Language."""

    identifier = "fireeye-helix"
    active = True
    index_field = "metaclass"
    nonTaxonomyField = "rawmsg"

    andToken = " "
    orToken = " OR "
    notToken = "NOT "
    subExpression = "(%s)"
    listExpression = "[%s]"
    listAndExpression = "&[%s]"
    listSeparator = ","
    valueExpression = "`%s`"
    nullExpression = "missing(%s)"
    notNullExpression = "has(%s)"
    mapExpression = "%s=%s"
    mapContainsExpression = "%s:%s"
    typedValueExpression = {
        SigmaRegularExpressionModifier: "/%s/",
        SigmaContainsModifier: "%s",
        SigmaStartswithModifier: "%s",
        SigmaEndswithModifier: "%s",
    }
    containsMatchFields = [index_field, nonTaxonomyField]
    exactMatchFields = ["eventid", "srcport", "dstport", "channel"]

    def __init__(self, *args, **kwargs):
        """Initialize field mappings"""
        super().__init__(*args, **kwargs)
        # Retrieve a list of fields explicitly mapped in the config so we can use "rawmsg" for unmapped fields
        fl = ["metaclass", "channel"]
        for item in self.sigmaconfig.fieldmappings.values():
            if item.target_type == list:
                fl.extend(item.target)
            else:
                fl.append(item.target)
        self.allowedFieldsList = list(set(fl))

    def generateMapItemNode(self, node):
        key, value = node
        # Default our expression to a contains match
        divinedExpression = self.mapContainsExpression
        # If a field defines what the expression would be, use this to lock it in despite what the value determines
        fieldForcedExpression = False
        # Special field handling
        if key not in self.allowedFieldsList:
            key = self.nonTaxonomyField
        if key in self.containsMatchFields:
            divinedExpression = self.mapContainsExpression
            fieldForcedExpression = True
        elif key in self.exactMatchFields:
            divinedExpression = self.mapExpression
            fieldForcedExpression = True

        # Always exact match integers
        if isinstance(value, int):
            if not fieldForcedExpression:
                divinedExpression = self.mapExpression
            return divinedExpression % (key, self.generateNode(value))
        elif isinstance(value, str):
            value, divinedExpression = self.parseStringValue(
                key, value, divinedExpression, fieldForcedExpression, False
            )
            return divinedExpression % (key, self.generateNode(value))
        elif isinstance(value, list):
            newList = []
            # Iterate over our list values to deal with wildcards in strings
            for _value in value:
                if isinstance(_value, str):
                    _value, divinedExpression = self.parseStringValue(
                        key, _value, divinedExpression, fieldForcedExpression, True
                    )
                    newList.append(_value)
                else:
                    newList.append(_value)
            return divinedExpression % (key, self.generateNode(newList))
        elif isinstance(value, SigmaTypeModifier):
            if isinstance(value, (SigmaStartswithModifier, SigmaEndswithModifier)):
                divinedExpression = self.mapContainsExpression
                # Strip prefix/suffix matching wildcards on rawmsg field searches
                if key == self.nonTaxonomyField and isinstance(value, str):
                    value.strip("*")
            elif isinstance(value, SigmaContainsModifier):
                divinedExpression = self.mapContainsExpression
            elif isinstance(value, SigmaRegularExpressionModifier):
                divinedExpression = self.mapContainsExpression
            return divinedExpression % (key, self.generateTypedValueNode(value))
        elif value is None:
            return self.nullExpression % (key,)
        else:
            raise TypeError(
                "Backend does not support map values of type " + str(type(value))
            )

    def generateNULLValueNode(self, node):
        # Don't generate null value nodes for fields we don't map
        if node.item == "rawmsg":
            return None
        else:
            return self.notNullExpression % (node.item)

    def generateNotNULLValueNode(self, node):
        # Don't generate not null value nodes for fields we don't map
        if node.item == "rawmsg":
            return None
        else:
            return self.nullExpression % (node.item)

    def parseStringValue(
        self, key, value, divinedExpression, fieldForcedExpression, isList
    ):
        # Raise a NotImplementedError if the query contains mid value wildcards for now
        # TODO figure out how to best handle this
        if "*" in value[1:-1]:
            raise NotImplementedError(
                "Backend does not support queries containing mid value wildcards."
            )
        # Strip balanced wildcards
        if value.startswith("*") and value.endswith("*"):
            value = value[1:-1]
            if not fieldForcedExpression:
                divinedExpression = self.mapContainsExpression
        # Prefix/suffix matches are "contains" operators
        elif value.startswith("*") or value.endswith("*"):
            # Strip wildcards from rawmsg matching because we don't know where in the log we're matching from
            if key == self.nonTaxonomyField:
                value = value.strip("*")
            if not fieldForcedExpression:
                divinedExpression = self.mapContainsExpression
        # If we have no indicators of this being a "contains" match for a string, use an exact match
        else:
            if not fieldForcedExpression and not isList:
                divinedExpression = self.mapExpression

        return value, divinedExpression



#!/usr/bin/env python3
# Merge a Sigma rule collection into full Sigma rules
# Copyright 2017 Thomas Patzke

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import argparse
import yaml

from sigma.parser.collection import SigmaCollectionParser

def main():
    argparser = argparse.ArgumentParser(description="Convert Sigma rules into SIEM signatures.")
    argparser.add_argument("input", help="Sigma input file")
    cmdargs = argparser.parse_args()

    try:
        f = open(cmdargs.input, "r")
    except IOError as e:
        print("Error while opening input file: %s" % str(e), file=sys.stderr)
        sys.exit(1)

    content = "".join(f.readlines())
    f.close()
    sc = SigmaCollectionParser(content)

    print(yaml.dump_all(sc, default_flow_style=False))

if __name__ == "__main__":
    main()
